import inspect

from numpy import real, imag, asarray

from bumps.parameter import Parameter, BaseParameter
from refl1d.material import SLD
from refl1d.model import Layer
from refl1d.magnetism import BaseMagnetism, Magnetism
from refl1d import util

class FunctionalProfile(Layer):
    """
    Generic profile function

    Parameters:

        *thickness* the thickness of the layer

        *interface* the roughness of the surface [not implemented]

        *profile* the profile function, suitably parameterized

        *tol* is the tolerance for considering values equal

        *magnetism* magnetic profile associated with the layer

        *name* is the layer name

    The profile function takes a depth vector *z* returns a density vector
    *rho*. For absorbing profiles, return complex vector *rho + irho*1j*.

    Fitting parameters are the available named arguments to the function.
    The first argument is a depth vector, which is the array of depths at
    which the profile is to be evaluated.  It is guaranteed to be increasing,
    with step size 2*z[0].

    Initial values for the function parameters can be given using name=value.
    These values can be scalars or fitting parameters.  The function will
    be called with the current parameter values as arguments.  The layer
    thickness can be computed as :func:`layer_thickness`.

    There is no mechanism for querying the larger profile to determine the
    value of the *rho* at the layer boundaries.  If needed, this information
    will have to be communicated through shared parameters.  For example::

        L1 = SLD('L1', rho=2.07)
        L3 = SLD('L3', rho=4)
        def linear(z, rhoL, rhoR):
            rho = z * (rhoR-rhoL)/(z[-1]-z[0]) + rhoL
            return rho
        profile = FunctionalProfile(100, 0, profile=linear,
                                    rhoL=L1.rho, rhoR=L3.rho)
        sample = L1 | profile | L3
    """
    RESERVED = ('thickness', 'interface', 'profile', 'tol', 'magnetism', 'name')

    # TODO: test that thickness(z) matches the thickness of the layer
    def __init__(self, thickness=0, interface=0, profile=None, tol=1e-3,
                 magnetism=None, name=None, **kw):
        if not name:
            name = profile.__name__
        if interface != 0:
            raise NotImplementedError("interface not yet supported")
        if profile is None:
            raise TypeError("Need profile")
        self.name = name
        self.thickness = Parameter.default(thickness, name=name+" thickness")
        self.interface = Parameter.default(interface, name=name+" interface")
        self.profile = profile
        self.tol = tol
        self.magnetism = magnetism

        # TODO: maybe make these lazy (and for magnetism below as well)
        rho_start = _LayerLimit(self, isend=False, isrho=True)
        irho_start = _LayerLimit(self, isend=False, isrho=False)
        rho_end = _LayerLimit(self, isend=True, isrho=True)
        irho_end = _LayerLimit(self, isend=True, isrho=False)
        self.start = SLD(name+" start", rho=rho_start, irho=irho_start)
        self.end = SLD(name+" end", rho=rho_end, irho=irho_end)

        self._parameters = _set_vars(self, name, profile, kw, self.RESERVED)

    def parameters(self):
        P = dict((k, getattr(self, k)) for k in self._parameters)
        P['thickness'] = self.thickness
        #P['interface'] = self.interface
        return P

    def render(self, probe, slabs):
        Pw, Pz = slabs.microslabs(self.thickness.value)
        if len(Pw) == 0:
            return
        #print kw
        # TODO: always return rho, irho from profile function
        # return value may be a constant for rho or irho
        phi = asarray(self.profile(Pz, **self._fpars()))
        if phi.shape != Pz.shape:
            raise TypeError("profile function '%s' did not return array phi(z)"
                            %self.profile.__name__)
        Pw, phi = util.merge_ends(Pw, phi, tol=self.tol)
        #P = M*phi + S*(1-phi)
        slabs.extend(rho=[real(phi)], irho=[imag(phi)], w=Pw)
        #slabs.interface(self.interface.value)

    def _fpars(self):
        kw = dict((k, getattr(self, k).value) for k in self._parameters)
        return  kw


class FunctionalMagnetism(BaseMagnetism):
    """
    Functional magnetism profile.

    Parameters:

        *profile* the profile function, suitably parameterized

        *tol* is the tolerance for considering values equal

        :class:`refl1d.magnetism.BaseMagnetism` parameters

    The profile function takes a depth vector *z* and returns a magnetism
    vector *rhoM*. For magnetic twist, return a pair of vectors *(rhoM, thetaM)*.
    Constants can be returned for *rhoM* or *thetaM*.

    See :class:`FunctionalProfile` for a description of the the profile
    function.
    """
    RESERVED = ('profile', 'tol', 'name', 'extent', 'dead_below', 'dead_above',
                'interface_below', 'interface_above')
    magnetic = True
    def __init__(self, profile=None, tol=1e-3, name=None, **kw):
        if not name:
            name = profile.__name__
        if profile is None:
            raise TypeError("Need profile")
        # strip magnetism keywords from list of keywords
        magkw = dict((a, kw.pop(a)) for a in set(self.RESERVED)&set(kw.keys()))
        BaseMagnetism.__init__(self, name=name, **magkw)
        self.profile = profile
        self.tol = tol

        self._parameters = _set_vars(self, name, profile, kw, self.RESERVED)
        rhoM_start = _MagnetismLimit(self, isend=False, isrhoM=True)
        rhoM_end = _MagnetismLimit(self, isend=True, isrhoM=True)
        thetaM_start = _MagnetismLimit(self, isend=False, isrhoM=False)
        thetaM_end = _MagnetismLimit(self, isend=True, isrhoM=False)
        self.start = Magnetism(rhoM=rhoM_start, thetaM=thetaM_start)
        self.end = Magnetism(rhoM=rhoM_end, thetaM=thetaM_end)
        self.anchor = None

    def set_anchor(self, stack, index):
        self.anchor = (stack, index)

    # TODO: is there a sane way of computing magnetism thickness in advance?
    def _calc_thickness(self):
        try:
            stack, index = self.anchor
        except:
            raise ValueError("\
Need layer.magnetism.set_anchor(stack, layer) to compute magnetic thickness.")

        stack, start = stack._lookup(index)
        total = 0
        for k in range(start, start+self.extent):
            total += stack[k].thickness.value
        total -= self.dead_below.value + self.dead_above.value
        return total

    def parameters(self):
        parameters = BaseMagnetism.parameters(self)
        parameters.update((k, getattr(self, k)) for k in self._parameters)
        return parameters

    def render(self, probe, slabs, thickness, anchor, sigma):
        Pw, Pz = slabs.microslabs(thickness)
        if len(Pw) == 0: return
        kw = dict((k, getattr(self, k).value) for k in self._parameters)
        P = self.profile(Pz, **kw)
        # TODO: always return rhoM, thetaM from profile function
        # rhoM or thetaM may be constant
        try:
            rhoM, thetaM = P
        except Exception:
            rhoM, thetaM = P, Pz*0
        rhoM, thetaM = [asarray(v) for v in (rhoM, thetaM)]
        if rhoM.shape != Pz.shape or thetaM.shape != Pz.shape:
            raise TypeError("profile function '%s' did not return array rhoM(z)"
                            %self.profile.__name__)
        P = rhoM + thetaM*0.001j  # combine rhoM/thetaM so they can be merged
        Pw, P = util.merge_ends(Pw, P, tol=self.tol)
        rhoM, thetaM = P.real, P.imag*1000  # split out rhoM,thetaM again
        slabs.add_magnetism(anchor=anchor,
                            w=Pw, rhoM=rhoM, thetaM=thetaM,
                            sigma=sigma)

    def _fpars(self):
        kw = dict((k, getattr(self, k).value) for k in self._parameters)
        return  kw

    def __repr__(self):
        return "FunctionalMagnetism(%s)"%self.name


def _set_vars(self, name, profile, kw, reserved):
    # Query profile function for the list of arguments
    vars = inspect.getargspec(profile)[0]
    #print "vars", vars
    if inspect.ismethod(profile):
        vars = vars[1:]  # Chop self
    vars = vars[1:]  # Chop z
    #print vars
    unused = [k for k in kw.keys() if k not in vars]
    if len(unused) > 0:
        raise TypeError("Profile got unexpected keyword argument '%s'"%unused[0])
    dups = [k for k in vars if k in reserved]
    if len(dups) > 0:
        raise TypeError("Profile has conflicting argument %r"%dups[0])
    for k in vars:
        kw.setdefault(k, 0)
    for k, v in kw.items():
        setattr(self, k, Parameter.default(v, name=name+" "+k))

    return vars

class _LayerLimit(BaseParameter):
    def __init__(self, flayer, isend=True, isrho=True):
        self.flayer = flayer
        self.isend = isend
        self.isrho = isrho
        self.name = (str(flayer)
                     + (".rho_" if isrho else ".irho_")
                     + ("end" if isend else "start"))

    def parameters(self):
        return None

    @property
    def value(self):
        z = asarray([0., self.flayer.thickness.value])
        P = self.flayer.profile(asarray(z), **self.flayer._fpars())
        index = 1 if self.isend else 0
        return real(P[index]) if self.isrho else imag(P[index])

    def __repr__(self):
        return repr(self.flayer) + self._tag

class _MagnetismLimit(BaseParameter):
    def __init__(self, flayer, isend=True, isrhoM=True):
        self.flayer = flayer
        self.isend = isend
        self.isrhoM = isrhoM
        self.name = (str(flayer)
                     + (".rhoM_" if isrhoM else ".thetaM_")
                     + ("end" if isend else "start"))

    def parameters(self):
        return None

    @property
    def value(self):
        zmax = self.flayer._calc_thickness()
        # TODO: fix interface
        # awkward: calculating mid-point so that we don't accidentally
        # interpret a two value return value as rhoM, thetaM instead of
        # rhoM, [0., 0.]
        z = asarray([0., 0.5*zmax, zmax])
        P = self.flayer.profile(z, **self.flayer._fpars())
        try:
            rhoM, thetaM = P   # Returns rhoM and thetaM
        except Exception:
            rhoM, thetaM = P, [0., 0., 0.]  # Returns rhoM only
        index = 2 if self.isend else 0
        return rhoM[index] if self.isrhoM else thetaM[index]

    def __repr__(self):
        return repr(self.flayer) + self._tag



"""
Reflectometry thickness interactor.
"""
from .config import thickness_color, pick_radius
from .interactor import BaseInteractor
from .util import clip

class ThicknessInteractor(BaseInteractor):
    """
    Control the size of the layers.
    """
    def __init__(self, profile):
        BaseInteractor.__init__(self, profile)
        self.markers  = []
        self.textmark = []
        self._show_labels = True
        self._show_boundaries = True

    def label_offsets(self):
        z = self.profile.boundary[1:-1]
        left = -20
        middle = [(a+b)/2. for a,b in zip(z[:-1],z[1:])]
        right = z[-1] + 20
        return [left]+middle+[right]

    def reset_markers(self):
        """
        Reset all markers.
        """
        self.clear_markers()
        ax = self.profile.axes

        # Add bars
        style = dict(linewidth=1, linestyle='-',
                     color=thickness_color, alpha=0.75,
                     pickradius=pick_radius,
                     visible=self._show_boundaries,
                     )
        self.markers = [ax.axvline(x=z, **style)
                        for z in self.profile.boundary[1:-1]]

        fittable = [self.profile.sample_layer(idx).thickness.fittable
                    for idx,_ in enumerate(self.markers)]
        fittable[0] = False # First interface is not fittable
        for f,m in zip(fittable,self.markers):
            if not f: m.set(linestyle='--', linewidth=1.25)
        self.connect_markers(m for f,m in zip(fittable,self.markers) if f)

        # Add labels
        offsets = self.label_offsets()
        labels = self.profile.sample_labels()
        style = dict(transform=self.profile.xcoords,
                     ha='left', va='bottom',
                     rotation=30, fontsize='small',
                     visible=self._show_labels,
                     )
        self.textmark = [ax.text(z,1,s,**style)
                         for z,s in zip(offsets,labels)]

    def clear_markers(self):
        """
        Remove interfaces and layer names from the graph.
        """
        BaseInteractor.clear_markers(self)
        for h in self.textmark:
            h.remove()
        self.textmark = []

    def update_markers(self):
        """
        Update the marker positions
        """
        label_offsets = self.label_offsets()

        for z,h in zip(self.profile.boundary[1:-1], self.markers):
            h.set_xdata([z,z])
        for z,h in zip(label_offsets, self.textmark):
            h.set_x(z)

    def save(self, ev):
        idx = self.markers.index(ev.artist)
        curr = self.profile.sample_layer(idx).thickness
        next = self.profile.sample_layer(idx+1).thickness \
               if idx<len(self.markers)-1 else None
        if not curr.fittable: curr = None
        if next is not None and not next.fittable: next = None

        self._idx = idx
        self._curr = curr
        self._next = next
        if curr is not None: self._curr_saved = curr.value
        if next is not None: self._next_saved = next.value

    def restore(self, ev):
        self._curr.value = self._curr_saved
        if self._next is not None:
            self._next.value = self._next_saved

    def drag(self, event):
        """
        Process move to a new position, making sure that the move is allowed.
        """
        if self._curr is None: return

        prev_offset = self.profile.boundary[self._idx]
        next_offset = self.profile.boundary[self._idx+2]

        # Limit the position according to parameter limits
        if event.shift and self._next is not None:
            # shifting interface between two layers
            lo_curr,hi_curr = self._curr.bounds.limits
            lo_next,hi_next = self._next.bounds.limits
            lo = min(lo_curr+prev_offset, next_offset-hi_next)
            hi = min(hi_curr+prev_offset, next_offset-lo_next)
        else:
            # resizing a layer
            lo_curr,hi_curr = self._curr.bounds.limits
            lo = lo_curr+prev_offset
            hi = hi_curr+prev_offset
        x = clip(event.xdata, lo, hi)

        # Set the current and next value based on offsets.  We set the next
        # in case the user is shifting the interface between two layers
        # instead of resizing the current layer.  We set it even when not
        # shifting in case the previous action was to shift, and the current
        # action triggers a restore.  We set the next before the current
        # so that if the two parameters are tied, the current takes
        # precedence.
        if self._next is not None:
            if event.shift:
                self._next.value = next_offset-x
            else:
                self._next.value = self._next_saved
        self._curr.value = x - prev_offset

        # Adjust x_offset such that the layer boundary will stay under the
        # curesor even if previous layers grow or shrink as a result of being
        # tied to the current layer thickness.
        self.profile._find_layer_boundaries()
        self.profile.x_offset -= self.profile.boundary[self._idx]-prev_offset



from __future__ import division
from math import log
import numpy
from .interactor import BaseInteractor
from .layer import MaterialInteractor
from .config import pick_radius
from .config import profile_color
from .config import vf_scale
from .util import setpar


#-----------------------------------------------------------------------
class PolymerBrushInteractor(BaseInteractor):
    """
    Interactor for tethered polymer model.
    """
    def __init__(self, profile, layer):
        super(PolymerBrushInteractor,self).__init__(profile)
        self.polymer = MaterialInteractor(profile,layer.polymer,
                                          range=(0,.3))
        self.solvent = MaterialInteractor(profile, layer.solvent,
                                          range=(0.7,1))
        self.layer = layer
        ax = profile.axes

        style = dict(linestyle = ':',
                     transform=profile.xcoords,
                     zorder=0,
                     color=profile_color,
                     visible=True,
                     )
        self.hprofile = ax.plot([],[], **style)[0]

        style = dict(linestyle='-',
                     transform=profile.xcoords,
                     zorder=4,
                     pickradius = pick_radius,
                     color = profile_color,
                     linewidth = 1,
                     visible=True,
                     )
        self.hphi = ax.plot([], [], **style)[0]
        self.hbase = ax.plot([],[], **style)[0]
        self.hlength = ax.plot([],[], **style)[0]

        style = dict(marker='s',
                     transform=profile.xcoords,
                     zorder=5,
                     pickradius = pick_radius,
                     color = profile_color,
                     alpha = 0.5,
                     markersize = 5,
                     visible=True,
                     )
        self.hpower = ax.plot([],[], **style)[0]
        self.hsigma = ax.plot([],[], **style)[0]
        #print "xcoords",id(profile.xcoords),"trans",id(self.hbase.get_transform())

        self.markers = [self.hphi,self.hbase,self.hlength,
                        self.hpower,self.hsigma]
        self.parameters = [layer.base_vf,layer.base,layer.length,
                           layer.power,layer.sigma]
        self.connect_markers(self.markers)

    def clear_markers(self):
        super(PolymerBrushInteractor,self).clear_markers()
        self.polymer.clear_markers()
        self.solvent.clear_markers()
        self.hprofile.remove()

    def update_markers(self):
        """
        Draw the widgets in their new positions.
        """
        self.polymer.update_markers()
        self.solvent.update_markers()

        n = self.profile.layer_num
        left = self.profile.boundary[n]
        layer = self.layer

        # TODO: want to use smae step size for profile as we do for
        # smooth profiles  Don't know how to access that value nicely.
        z = numpy.arange(0,layer.thickness.value,0.1)
        vf = layer.profile(z)
        self.hprofile.set_data(z+left, vf*vf_scale)

        # at z midway between l1 and l2
        #     y = phi * 0.75**p
        phi = layer.base_vf.value*vf_scale/100
        phi_power = phi*0.75**layer.power.value
        L0 = left + layer.base.value
        L1 = L0 + layer.length.value
        Lpower = (L0+L1)/2
        Lsigma = L1 + layer.sigma.value

        #print "L0,L1,Lpower,Lsigma",L0,L1,Lpower,Lsigma,phi,phi_power
        self.hphi.set_data( (left,L0), (phi,phi) )
        self.hbase.set_data( (L0,L0),  (0,phi) )
        self.hlength.set_data( (L1,L1), (0, phi) )
        self.hpower.set_data( (Lpower,), (phi_power,) )
        self.hsigma.set_data( (Lsigma,), (0,) )

    def drag(self, ev):
        """
        Update the model with the new widget position.
        """
        table = dict(zip(self.markers,self.parameters))
        par = table[ev.artist]
        n = self.profile.layer_num
        left = self.profile.boundary[n]
        right = self.profile.boundary[n+1]
        # TODO: Keeping individual interactors within the layer does not
        # necessarily keep the whole interactor in the layer.
        z = max(left,min(right,ev.xdata))
        if ev.artist == self.hphi:
            #print ev.ydata, self.profile.xcoords.inverted().transform([(ev.x,ev.y)])[0][1]
            setpar(par, ev.ydata*100/vf_scale)
            #print "phi",par.value,ev.ydata*100/vf_scale
        elif ev.artist == self.hbase:
            offset = left
            setpar(par, z-offset)
        elif ev.artist == self.hlength:
            offset = left + self.layer.base.value
            setpar(par, z-offset)
        elif ev.artist == self.hsigma:
            offset = left + self.layer.base.value + self.layer.length.value
            setpar(par, z-offset)
        elif ev.artist == self.hpower:
            phi = self.layer.base_vf.value*vf_scale/100
            phi_power = log(ev.ydata/phi)/log(0.75) if ev.ydata > 0 else 100
            setpar(par, phi_power)


    def save(self, ev):
        """
        Save the current state of the model represented by the widget.
        """
        table = dict(zip(self.markers,self.parameters))
        par = table[ev.artist]
        self._save_par = par
        self._save_val = par.value

    def restore(self, ev):
        """
        Restore the widget and model to the saved state.
        """
        self._save_par.value = self._save_val



"""
Layer interactor.
"""
from __future__ import division
from bumps.parameter import Parameter
from refl1d.material import SLD
from .interactor import BaseInteractor
from .config import pick_radius
from .config import rho_color, rhoI_color, disabled_color
from .util import setpar


class MaterialInteractor(BaseInteractor):
    def __init__(self, profile, material, range=(0,1)): #@ReservedAssignment
        super(MaterialInteractor,self).__init__(profile)
        self.material = material
        self.range = range

        if isinstance(material,SLD):
            enabled = True
            self._rho = material.rho
            self._rhoI = material.irho
            self._rho_scale = 1
            self._rhoI_scale = 1
        elif (hasattr(material,'density')
              and isinstance(material.density,Parameter)):
            enabled = True
            rho,rhoI = material.sld(profile.experiment.probe)
            self._rho = material.density
            self._rhoI = material.density
            self._rho_scale = 1/ (rho / material.density.value)
            self._rhoI_scale = 1/ (rhoI / material.density.value)
        else:
            self._rho = self._rhoI = None
            enabled = False

        style = dict(linestyle='--',
                     linewidth  = 2,
                     pickradius = pick_radius,
                     )
        ax = profile.axes
        rho = ax.plot([], [], label=material.name+" rho",
                      color=rho_color if enabled else disabled_color,
                      zorder=6, **style)[0]
        rhoI = ax.plot([], [], label=material.name+" rhoI",
                      color=rhoI_color if enabled else disabled_color,
                      zorder=5, **style)[0]

        self.markers = [rho,rhoI]
        if enabled:
            self.connect_markers(self.markers)

    def update_markers(self):
        z = self.profile.boundary[1:-1]
        n = self.profile.layer_num
        #if n is None: return
        if n == 0:
            left,right = -20,0
        elif n >= len(z):
            left,right = z[-1],z[-1]+20
        else:
            delta = z[n] - z[n-1]
            left,right = self.range[0]*delta+z[n-1], self.range[1]*delta+z[n-1]

        rho,rhoI = self.material.sld(self.profile.experiment.probe)
        self.markers[0].set_data( (left,right), (rho,rho) )
        self.markers[1].set_data( (left,right), (rhoI,rhoI) )

    def save(self, ev):
        if self._rho is not None: self._rho_save = self._rho.value
        if self._rhoI is not None: self._rhoI_save = self._rhoI.value

    def restore(self, ev):
        if self._rho is not None: self._rho.value = self._rho_save
        if self._rhoI is not None: self._rhoI.value = self._rhoI_save

    def drag(self, ev):
        if ev.artist == self.markers[0]:
            setpar(self._rho, ev.ydata * self._rho_scale)
        else:
            setpar(self._rhoI, ev.ydata * self._rhoI_scale)

class SlabInteractor(BaseInteractor):
    def __init__(self, profile, layer):
        super(SlabInteractor,self).__init__(profile)
        self.material = MaterialInteractor(profile,layer.material)
    def update_markers(self):
        self.material.update_markers()
    def clear_markers(self):
        self.material.clear_markers()


# -------------------------------------------------------------------
class NoInteractor(BaseInteractor):
    """
    Null Interactor for undefined layers.
    """
    def __init__(self, profile, layer):
        super(NoInteractor,self).__init__(profile)
    # TODO: turn layer.parameters() into sliders



# Some global variables for interactor

pick_radius = 5
layer_hysteresis = 4

thickness_color = 'black'
interface_color = 'black'
disabled_color  = 'gray'
active_color    = 'red'
rho_color       = 'black'
rhoI_color      = 'green'
rhoM_color      = 'blue'
thetaM_color    = 'orange'
profile_color   = 'cyan'
title_color     = 'red'

vf_scale = 0.4



"""
interface  interactor.
"""

from .config import interface_color, pick_radius
from .interactor import BaseInteractor
from .util import clip, setpar

MAX_ROUGH=1

class InterfaceInteractor(BaseInteractor):
    """
    Control the roughness of the layers.
    """
    def __init__(self, profile):
        BaseInteractor.__init__(self, profile)
        ax = profile.axes

        # markers for roughness
        style = dict(linestyle  = '',
                     transform  = profile.xcoords,
                     marker     = 's', #square
                     markersize = 7,
                     color      = interface_color,
                     alpha      = 0.6,
                     pickradius = pick_radius,
                     zorder     = 8, #Prefer this to other lines
                     visible    = False,
                     )
        self.markers=[ax.plot([0],[0.05], label=label, **style)[0]
                      for label in ('interface marker L','interface marker R')]

        # lines for roughness
        style = dict(linestyle = '-',
                     transform = profile.xcoords,
                     marker    = '',
                     color     = interface_color,
                     visible   = False)
        self.lines=[ax.plot([0,0],[0.05,0.05], label=label, **style)[0]
                    for label in ('interface line L','interface line R')]

        self.connect_markers(self.markers)
        self._left = self._right = None


    def set_layer(self):
        """
        Move markers to the new layer
        """
        n = self.profile.layer_num
        z = self.profile.boundary[1:-1]
        show_left = n is not None and n>0
        show_right = n is not None and n < len(z)
        if show_left:
            self._left = self.profile.sample_layer(n-1).interface
        else:
            self._left = None
        if show_right:
            self._right = self.profile.sample_layer(n).interface
        else:
            self._right = None
        #self.update_markers()


    def update_markers(self):
        """
        Draw the new roughness on the graph.
        """
        n = self.profile.layer_num
        z = self.profile.boundary[1:-1]
        show_left = self._left is not None
        show_right = self._right is not None

        self.markers[0].set(visible=show_left)
        self.lines[0].set(visible=show_left)
        if show_left:
            self.markers[0].set_xdata([z[n-1]+self._left.value])
            self.lines[0].set_xdata([z[n-1], z[n-1]+self._left.value])

        self.markers[1].set(visible=show_right)
        self.lines[1].set(visible=show_right)
        if show_right:
            self.markers[1].set_xdata([z[n]-self._right.value])
            self.lines[1].set_xdata([z[n],z[n]-self._right.value])

    def clear_markers(self):
        """
        Remove interface markers from the graph.
        """
        BaseInteractor.clear_markers(self)
        for h in self.lines:
            h.remove()
        self.lines = []


    def save(self, event):
        """
        Remember the interface for this layer and the next so that we
        can restore on Esc.
        """

        if self._left is not None:
            self._left_value = self._left.value
        if self._right is not None:
            self._right_value = self._right.value

    def restore(self, event):
        """
        Restore the roughness for this layer.
        """
        if self._left is not None:
            self._left.value = self._left_value
        if self._right is not None:
            self._right.value = self._right_value

    def drag(self, event):
        """
        Process move to a new position, making sure that the move is allowed.
        """
        z = self.profile.boundary
        n = self.profile.layer_num
        side = self.markers.index(event.artist)
        if side == 0:  # Left
            limit = min(z[n]-z[n-1], z[n+1]-z[n])
            v = clip(event.xdata-z[n], 0, limit/MAX_ROUGH)
            setpar(self._left, v)

        else:  # Right
            limit = min(z[n+1]-z[n], z[n+2]-z[n+1])
            v = clip(z[n+1] - event.xdata, 0, limit/MAX_ROUGH)
            setpar(self._right, v)

        #self.update_markers()



"""
Reflectivity plugin for fitting GUI.

Note that the fitting infrastructure is still heavily tied to the reflectivity
modeling program, and this represents only the first tiny steps to separating
the two.
"""

__all__ = ["data_view", "model_view", "new_model", "calc_errors", "show_errors"]

import numpy as np

from . import names as refl
from .errors import calc_errors, show_errors

# These are names used by the driver
def data_view():
    from .view.data_view import DataView
    return DataView

def model_view():
    from .view.model_view import ModelView
    return ModelView

def load_model(filename):
    # TODO: bumps plugin api needs to allow options for loader
    options = None
    if (filename.endswith('.so') or filename.endswith('.dll')
            or filename.endswith('.dyld')):
        from . import garefl
        problem = garefl.load(filename)
        return problem
    elif filename.endswith('.staj') or filename.endswith('.sta'):
        from .stajconvert import load_mlayer
        return refl.FitProblem(load_mlayer(filename))
        #fit_all(problem.fitness, pmp=20)
    elif filename.endswith('.zip'):
        from bumps.fitproblem import load_problem
        # Note: bumps.vfs.vfs_init() must be called very early
        try:
            from bumps.vfs import ZipFS
        except ImportError:
            raise NotImplementedError("need newer bumps to load model from zip")
        with ZipFS(filename) as zf:
            for f in zf.filelist:
                if f.filename.endswith('.py'):
                    return load_problem(f.filename, options=options)
    else:
        return None

def new_model():
    stack = refl.silicon(0, 10) | refl.air
    instrument = refl.NCNR.NG1()
    probe = instrument.probe(T=np.linspace(0, 5, 200),
                             Tlo=0.2, slits_at_Tlo=2)
    M = refl.Experiment(sample=stack, probe=probe)
    problem = refl.FitProblem(M)
    return problem



"""
Common materials in reflectometry experiments along with densities.

By name::

    air, water, heavywater, lightheavywater, silicon, sapphire, gold
    permalloy

By formula::

    H2O, D2O, DHO, Si, Al2O3, Au, Ni8Fe2

If you want to adjust the density you will need to make your own copy of
these materials.  For example, for permalloy::

    >>> NiFe=Material(permalloy.formula, density=permalloy.bulk_density)
    >>> NiFe.density.pmp(10)  # Let density vary by 10% from bulk value
    Parameter(permalloy density)
"""
import periodictable
from .material import Vacuum, Material

__all__ = ['air',
           'water', 'H2O', 'heavywater', 'D2O', 'lightheavywater', 'DHO',
           'silicon', 'Si', 'sapphire', 'Al2O3', 'gold', 'Au',
           'permalloy', 'Ni8Fe2'
          ]

# Set the bulk density of heavy water assuming it has the same packing
# fraction as water, but with heavier H[2] substituted for H[1].
rho_D2O = periodictable.formula('D2O').mass/periodictable.formula('H2O').mass
rho_DHO = periodictable.formula('DHO').mass/periodictable.formula('H2O').mass

air = Vacuum()
water = H2O = Material('H2O', density=1, name='water')
heavywater = D2O = Material('D2O', density=rho_D2O)
lightheavywater = DHO = Material('DHO', density=rho_DHO)
silicon = Si = Material('Si')
sapphire = Al2O3 = Material('Al2O3', density=3.965, name='sapphire')
gold = Au = Material('Au', name='gold')
permalloy = Ni8Fe2 = Material('Ni8Fe2', density=8.692, name='permalloy')



"""
Basic reflectometry calculations

Slab model reflectivity calculator with optional absorption and roughness.
The function reflectivity_amplitude returns the complex waveform.
Slab model with supporting magnetic scattering.  The function
magnetic_reflectivity returns the complex reflection for the four
spin polarization cross sections [++, +-, -+, --].  The function
unpolarized_magnetic returns the expected magnitude for a measurement
of the magnetic scattering using an unpolarized beam.
"""
from __future__ import print_function

from six.moves import reduce

#__doc__ = "Fundamental reflectivity calculations"
__author__ = "Paul Kienzle"
__all__ = ['reflectivity', 'reflectivity_amplitude',
           'magnetic_reflectivity', 'magnetic_amplitude',
           'unpolarized_magnetic', 'convolve',
          ]

import numpy as np
from numpy import pi, sin, cos, conj, radians
# delay load so doc build doesn't require compilation
#from . import reflmodule

def _dense(x, dtype='d'):
    return np.ascontiguousarray(x, dtype)


def reflectivity(*args, **kw):
    """
    Calculate reflectivity $|r(k_z)|^2$ from slab model.

    :Parameters :
        *depth* : float[N] | |Ang|
            Thickness of the individual layers (incident and substrate
            depths are ignored)
        *sigma* : float OR float[N-1] | |Ang|
            Interface roughness between the current layer and the next.
            The final layer is ignored.  This may be a scalar for fixed
            roughness on every layer, or None if there is no roughness.
        *rho*, *irho* : float[N] OR float[N, K] | |1e-6/Ang^2|
            Real and imaginary scattering length density.  Use multiple
            columns when you have kz-dependent scattering length densities,
            and set rho_offset to select the appropriate one.  Data should
            be stored in column order.
        *kz* : float[M] | |1/Ang|
            Points at which to evaluate the reflectivity
        *rho_index* : integer[M]
            *rho* and *irho* columns to use for the various kz.

    :Returns:
        *R* | float[M]
            Reflectivity magnitude.

    This function does not compute any instrument resolution corrections.
    """
    r = reflectivity_amplitude(*args, **kw)
    return (r*conj(r)).real

def reflectivity_amplitude(kz=None,
                           depth=None,
                           rho=None,
                           irho=0,
                           sigma=0,
                           rho_index=None,
                          ):
    r"""
    Calculate reflectivity amplitude $r(k_z)$ from slab model.

    :Parameters :
        *depth* : float[N] | |Ang|
            Thickness of the individual layers (incident and substrate
            depths are ignored)
        *sigma* = 0 : float OR float[N-1] | |Ang|
            Interface roughness between the current layer and the next.
            The final layer is ignored.  This may be a scalar for fixed
            roughness on every layer, or None if there is no roughness.
        *rho*, *irho* = 0: float[N] OR float[N, K] | |1e-6/Ang^2|
            Real and imaginary scattering length density.  Use multiple
            columns when you have kz-dependent scattering length densities,
            and set *rho_index* to select amongst them.  Data should be
            stored in column order.
        *kz* : float[M] | |1/Ang|
            Points at which to evaluate the reflectivity
        *rho_index* = 0 : integer[M]
            *rho* and *irho* columns to use for the various kz.

    :Returns:
        *r* | complex[M]
            Complex reflectivity waveform.

    This function does not compute any instrument resolution corrections.
    """
    from . import reflmodule

    kz = _dense(kz, 'd')
    if rho_index is None:
        rho_index = np.zeros(kz.shape, 'i')
    else:
        rho_index = _dense(rho_index, 'i')

    depth = _dense(depth, 'd')
    if np.isscalar(sigma):
        sigma = sigma*np.ones(len(depth)-1, 'd')
    else:
        sigma = _dense(sigma, 'd')
    rho = _dense(rho, 'd')
    if np.isscalar(irho):
        irho = irho * np.ones_like(rho)
    else:
        irho = _dense(irho, 'd')


    #print(irho.shape, irho[:,0], irho[:,-1])
    irho[irho < 0] = 0.
    #print depth.shape, rho.shape, irho.shape, sigma.shape
    #print depth.dtype, rho.dtype, irho.dtype, sigma.dtype
    r = np.empty(kz.shape, 'D')
    #print "amplitude", depth, rho, kz, rho_index
    #print depth.shape, sigma.shape, rho.shape, irho.shape, kz.shape
    reflmodule._reflectivity_amplitude(depth, sigma, rho, irho, kz,
                                       rho_index, r)
    return r


def magnetic_reflectivity(*args, **kw):
    """
    Magnetic reflectivity for slab models.

    Returns the expected values for the four polarization cross
    sections (++, +-, -+, --).
    Return reflectivity R^2 from slab model with sharp interfaces.
    returns reflectivities.

    The parameters are as follows:

    kz (|1/Ang|)
        points at which to evaluate the reflectivity
    depth (|Ang|)
        thickness of the individual layers (incident and substrate
        depths are ignored)
    rho (|1e-6/Ang^2|)
        Scattering length density.
    irho (|1e-6/Ang^2|)
        absorption. Defaults to 0.
    rho_m (microNb)
        Magnetic scattering length density correction.
    theta_m (degrees)
        Angle of the magnetism within the layer.
    sigma (|Ang|)
        Interface roughness between the current layer and the next.
        The final layer is ignored.  This may be a scalar for fixed
        roughness on every layer, or None if there is no roughness.
    wavelength (|Ang|)
        Incident wavelength (only affects absorption).  May be a vector.
        Defaults to 1.
    Aguide (degrees)
        Angle of the guide field; -90 is the usual case

    This function does not compute any instrument resolution corrections.
    Interface diffusion, if present, uses the Nevot-Croce approximation.

    Use magnetic_amplitude to return the complex waveform.
    """
    r = magnetic_amplitude(*args, **kw)
    return [(z*z.conj()).real for z in r]

def unpolarized_magnetic(*args, **kw):
    """
    Returns the average of magnetic reflectivity for all cross-sections.

    See :class:`magnetic_reflectivity <refl1d.reflectivity.magnetic_reflectivity>` for details.
    """
    return reduce(np.add, magnetic_reflectivity(*args, **kw))/2.

B2SLD = 2.31604654  # Scattering factor for B field 1e-6/

def magnetic_amplitude(kz,
                       depth,
                       rho,
                       irho=0,
                       rhoM=0,
                       thetaM=0,
                       sigma=0,
                       Aguide=-90,
                       H=0,
                       rho_index=None,
                      ):
    """
    Returns the complex magnetic reflectivity waveform.

    See :class:`magnetic_reflectivity <refl1d.reflectivity.magnetic_reflectivity>` for details.
    """
    from . import reflmodule

    kz = _dense(kz, 'd')
    if rho_index is None:
        rho_index = np.zeros(kz.shape, 'i')
    else:
        rho_index = _dense(rho_index, 'i')
    n = len(depth)
    if np.isscalar(irho):
        irho = irho*np.ones(n, 'd')
    if np.isscalar(rhoM):
        rhoM = rhoM*np.ones(n, 'd')
    if np.isscalar(thetaM):
        thetaM = thetaM*np.ones(n, 'd')
    if np.isscalar(sigma):
        sigma = sigma*np.ones(n-1, 'd')

    #kz = -kz
    #depth, rho, irho, sigma, rhoM, thetaM = [v[::-1] for v in (depth, rho, irho, sigma, rhoM, thetaM)]
    depth, rho, irho, sigma = [_dense(a, 'd') for a in (depth, rho, irho, sigma)]
    #np.set_printoptions(linewidth=1000)
    #print(np.vstack((depth, np.hstack((sigma, np.nan)), rho, irho, rhoM, thetaM)).T)

    sld_b, u1, u3 = calculate_u1_u3(H, rhoM, thetaM, Aguide)

    R1, R2, R3, R4 = [np.empty(kz.shape, 'D') for pol in (1, 2, 3, 4)]
    reflmodule._magnetic_amplitude(depth, sigma, rho, irho,
                                   sld_b, u1, u3, Aguide, kz, rho_index,
                                   R1, R2, R3, R4)
    return R1, R2, R3, R4

def calculate_u1_u3(H, rhoM, thetaM, Aguide):
    from . import reflmodule

    rhoM, thetaM = _dense(rhoM, 'd'), _dense(np.radians(thetaM), 'd')
    n = len(rhoM)
    u1, u3 = np.empty(n, 'D'), np.empty(n, 'D')
    sld_b = np.empty(n, 'd')
    reflmodule._calculate_u1_u3(H, rhoM, thetaM, Aguide, sld_b, u1, u3)

    return sld_b, u1, u3

def calculate_u1_u3_py(H, rhoM, thetaM, Aguide):
    rotate_M = True

    EPS = np.finfo('f').tiny # not 1e-20 # epsilon offset for divisions.
    thetaM = radians(thetaM)
    phiH = radians(Aguide - 270.0)
    thetaH = np.pi/2.0 # by convention, H is in y-z plane so theta = pi/2

    sld_h = B2SLD * H
    sld_m_x = rhoM * np.cos(thetaM)
    sld_m_y = rhoM * np.sin(thetaM)
    sld_m_z = 0.0 # by Maxwell's equations, H_demag = mz so we'll just cancel it here
    # The purpose of AGUIDE is to rotate the z-axis of the sample coordinate
    # system so that it is aligned with the quantization axis z, defined to be
    # the direction of the magnetic field outside the sample.
    if rotate_M:
        # rotate the M vector instead of the transfer matrix!
        # First, rotate the M vector about the x axis:
        new_my = sld_m_z * sin(radians(Aguide)) + sld_m_y * cos(radians(Aguide))
        new_mz = sld_m_z * cos(radians(Aguide)) - sld_m_y * sin(radians(Aguide))
        sld_m_y, sld_m_z = new_my, new_mz
        sld_h_x = sld_h_y = 0.0
        sld_h_z = sld_h
        # Then, don't rotate the transfer matrix
        Aguide = 0.0
    else:
        sld_h_x = sld_h * np.cos(thetaH) # zero
        sld_h_y = sld_h * np.sin(thetaH) * np.cos(phiH)
        sld_h_z = sld_h * np.sin(thetaH) * np.sin(phiH)

    sld_b_x = sld_h_x + sld_m_x
    sld_b_y = sld_h_y + sld_m_y
    sld_b_z = sld_h_z + sld_m_z

    # avoid divide-by-zero:
    sld_b_x += EPS * (sld_b_x == 0)
    sld_b_y += EPS * (sld_b_y == 0)

    # add epsilon to y, to avoid divide by zero errors?
    sld_b = np.sqrt(sld_b_x**2 + sld_b_y**2 + sld_b_z**2)
    u1_num = (+sld_b + sld_b_x + 1j*sld_b_y - sld_b_z)
    u1_den = (+sld_b + sld_b_x - 1j*sld_b_y + sld_b_z)
    u3_num = (-sld_b + sld_b_x + 1j*sld_b_y - sld_b_z)
    u3_den = (-sld_b + sld_b_x - 1j*sld_b_y + sld_b_z)

    u1 = u1_num/u1_den
    u3 = u3_num/u3_den
    #print "u1", u1
    #print "u3", u3
    return sld_b, u1, u3

def convolve(xi, yi, x, dx):
    """
    Apply x-dependent gaussian resolution to the theory.

    Returns convolution y[k] of width dx[k] at points x[k].

    The theory function is a piece-wise linear spline which does not need to
    be uniformly sampled.  The theory calculation points *xi* should be dense
    enough to capture the "wiggle" in the theory function, and should extend
    beyond the ends of the data measurement points *x*. Convolution at the
    tails is truncated and normalized to area of overlap between the resolution
    function in case the theory does not extend far enough.
    """
    from . import reflmodule

    x = _dense(x)
    y = np.empty_like(x)
    reflmodule.convolve(_dense(xi), _dense(yi), x, _dense(dx), y)
    return y


def convolve_sampled(xi, yi, xp, yp, x, dx):
    """
    Apply x-dependent arbitrary resolution function to the theory.

    Returns convolution y[k] of width dx[k] at points x[k].

    Like :func:`convolve`, the theory *(xi, yi)* is represented as a
    piece-wise linear spline which should extend beyond the data
    measurement points *x*.  Instead of a gaussian resolution function,
    resolution *(xp, yp)* is also represented as a piece-wise linear
    spline.
    """
    from . import reflmodule

    x = _dense(x)
    y = np.empty_like(x)
    reflmodule.convolve_sampled(_dense(xi), _dense(yi), _dense(xp), _dense(yp),
                                x, _dense(dx), y)
    return y


def test_convolve_sampled():
    x = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    y = [1, 3, 1, 2, 1, 3, 1, 2, 1, 3]
    xp = [-1, 0, 1, 2, 3]
    yp = [1, 4, 3, 2, 1]
    _check_convolution("aligned", x, y, xp, yp, dx=1)
    _check_convolution("unaligned", x, y, _dense(xp) - 0.2000003, yp, dx=1)
    _check_convolution("wide", x, y, xp, yp, dx=2)
    _check_convolution("super wide", x, y, xp, yp, dx=10)


def _check_convolution(name, x, y, xp, yp, dx):
    ystar = convolve_sampled(x, y, xp, yp, x, dx=np.ones_like(x) * dx)

    xp = np.array(xp) * dx
    step = 0.0001
    xpfine = np.arange(xp[0], xp[-1] + step / 10, step)
    ypfine = np.interp(xpfine, xp, yp)
    # make sure xfine is wide enough by adding a couple of extra steps
    # at the end
    xfine = np.arange(x[0] + xpfine[0], x[-1] + xpfine[-1] + 2 * step, step)
    yfine = np.interp(xfine, x, y, left=0, right=0)
    pidx = np.searchsorted(xfine, np.array(x) + xp[0])
    left, right = np.searchsorted(xfine, [x[0], x[-1]])

    conv = []
    for pi in pidx:
        norm_start = max(0, left - pi)
        norm_end = min(len(xpfine), right - pi)
        norm = step * np.sum(ypfine[norm_start:norm_end])
        conv.append(
            step * np.sum(ypfine * yfine[pi:pi + len(xpfine)]) / norm)

    #print("checking convolution %s"%(name, ))
    #print(" ".join("%7.4f"%yi for yi in ystar))
    #print(" ".join("%7.4f"%yi for yi in conv))
    assert all(abs(yi - fi) < 0.0005 for (yi, fi) in zip(ystar, conv))



# This program is in the public domain
# Author: Paul Kienzle
"""
Refl1D: Specular Reflectometry Modeling and Fitting Software

This package provides tools for modeling a variety of systems from
simple slabs to complex systems with a mixture models involving
smoothly varying layers.

X-ray and neutron and polarized neutron data can be loaded and the model
parameters adjusted to achieve the best fit.

A graphical interface allows direct manipulation of the model profiles.

See http://www.reflectometry.org/danse/reflectometry for online manuals.
"""

__version__ = "0.8.10"



"""
Visual representation of model uncertainty.

For reflectivity models, this aligns and plots a set of profiles chosen
from the parameter uncertainty distribution, and plots the distribution
of the residual values.

Use *run_errors* in a model file to reload the results of a batch DREAM fit.
"""
from __future__ import print_function

__all__ = ['reload_errors', 'run_errors',
           'calc_errors', 'align_profiles',
           'show_errors', 'show_profiles', 'show_residuals',
          ]

import numpy as np
from bumps.plotutil import next_color, dhsv, plot_quantiles, form_quantiles
from bumps.errplot import reload_errors

from .util import asbytes

#_CONTOURS = (68, 95, 100)
#_CONTOURS = (57, 68, 84, 95, 100)
_CONTOURS = (68, 95)

def run_errors(**kw):
    """
    Argument parser for generating error plots from models.

    The model directory should contain a fake model align.py with:

        from refl1d.errors import run_errors
        run_errors(model="", store="", align='auto')

    If you are using the command line then you should be able to type the
    following at the command prompt to generate the plots:

        $ refl1d align.py <model>.py <store> [<align>] [0|1|2|n]

    If you are using the GUI, you will have to set model, store and
    align directly in align.py each time you run.

    Align is either auto for the current behaviour, or it is an interface
    number. You can align on the center of a layer by adding 0.5 to the
    interface number. You can count interfaces from the surface by prefixing
    with R.  For example, 0 is the substrate interface, R1 is the surface
    interface, 2.5 is the the middle of layer 2 above the substrate.

    You can plot the profiles and residuals on one plot by setting plots to 1,
    on two separate plots by setting plots to 2, or each curve on its own
    plot by setting plots to n. Plots are saved in <store>/<model>-err#.png.
    If plots is 0, then no plots are created.

    Additional parameters include:

        *nshown*, *random* :

            see :func:`bumps.errplot.calc_errors_from_state`

        *contours*, *npoints*, *plots*, *save* :

            see :func:`show_errors`
    """
    import os
    import sys
    import matplotlib.pyplot as plt

    load = {'model': None, 'store': None, 'nshown': 50, 'random': True}
    show = {'align': 'auto', 'plots': 2,
            'contours': _CONTOURS, 'npoints': 400,
            'save': None}

    for k, v in kw.items():
        if k in load:
            load[k] = v
        elif k in show:
            show[k] = v
        else:
            raise TypeError("Unknown argument "+k)

    if len(sys.argv) > 2:
        load['model'], load['store'] = sys.argv[1], sys.argv[2]

        align_str = sys.argv[3] if len(sys.argv) >= 4 else '0'
        if align_str[0] == 'R':
            align_str = '-'+align_str[1:]
        show['align'] = float(align_str) if align_str != 'auto' else align_str
        plots_str = sys.argv[4] if len(sys.argv) >= 5 else '2'
        show['plots'] = int(plots_str) if plots_str != 'n' else plots_str
        #print align, align_str, len(sys.argv), sys.argv

    if not load['store'] or not load['model']:
        _usage()
    if show['save'] is None:
        show['save'] = os.path.join(load['store'], load['model'][:-3])

    print("loading... this may take awhile")
    errors = reload_errors(**load)
    print("showing...")
    show_errors(errors, **show)
    plt.show()
    raise KeyboardInterrupt()  # Force refl1d to terminate

def _usage():
    print(run_errors.__doc__)
    raise RuntimeError()


# TODO: we should just keep a certain number of evaluations as a matter of
#       course during sampling rather than recomputing them after the fact.
# TODO: want similar code for covariance matrix based forward analysis
# TODO: need to delegate accumulation of models and plotting to Fitness
# TODO: move run_errors somewhere more appropriate, like cli.py

def calc_errors(problem, points):
    """
    Align the sample profiles and compute the residual difference from the
    measured reflectivity for a set of points.

    The points should be sampled from the posterior probability
    distribution computed from MCMC, bootstrapping or sampled from
    the error ellipse calculated at the minimum.

    Each of the returned arguments is a dictionary mapping model number to
    error sample data as follows:

    Returns (profiles, slabs, Q, residuals).

    *profiles*

        Arrays of (z, rho, irho) for non-magnetic models or arrays
        of (z, rho, irho, rhoM, thetaM) for magnetic models.  There
        will be one set of arrays returned per error sample.

    *slabs*

        Array of slab thickness for the layers in the models.  There
        will be one array returned per error sample.  Using slab thickness,
        profiles can be aligned on interface boundaries and layer centers.

    *Q*

        Array of Q values for the data points in the model.  The data
        points are the same for all error samples, so only one Q array
        is needed per model.

    *residuals*

        Array of (theory-data)/uncertainty for each data point in
        the measurement.  There will be one array returned per error sample.
    """
    # Grab the individual samples
    if hasattr(problem, 'models'):
        models = [m.fitness for m in problem.models]
    else:
        models = [problem.fitness]

    experiments = []
    for m in models:
        if hasattr(m, 'parts'):
            experiments.extend(m.parts)
        else:
            experiments.append(m)
    #probes = []
    #for m in experiments:
    #    if hasattr(m.probe, 'probes'):
    #        probes.extend(m.probe.probes)
    #    elif hasattr(m.probe, 'xs'):
    #        probes.extend([p for p in m.probe if p])
    #    else:
    #        probes.append(p)

    # Find Q
    def residQ(m):
        if m.probe.polarized:
            return np.hstack([xs.Q for xs in m.probe.xs if xs is not None])
        else:
            return m.probe.Q
    Q = dict((m, residQ(m)) for m in experiments)

    profiles = dict((m, []) for m in experiments)
    residuals = dict((m, []) for m in experiments)
    slabs = dict((m, []) for m in experiments)
    def record_point():
        problem.chisq() # Force reflectivity recalculation
        for m in experiments:
            D = m.residuals()
            residuals[m].append(D+0)
            slabs_i = [L.thickness.value for L in m.sample[1:-1]]
            slabs[m].append(np.array(slabs_i))
            if m.ismagnetic:
                z, rho, irho, rhoM, thetaM = m.magnetic_step_profile()
                profiles[m].append((z+0, rho+0, irho+0, rhoM+0, thetaM+0))
            else:
                z, rho, irho = m.smooth_profile()
                profiles[m].append((z+0, rho+0, irho+0))
    record_point() # Put best at slot 0, no alignment
    for p in points:
        problem.setp(p)
        record_point()

    # Turn residuals into arrays
    residuals = dict((k, np.asarray(v).T) for k, v in residuals.items())
    return profiles, slabs, Q, residuals

def align_profiles(profiles, slabs, align):
    """
    Align profiles for each sample
    """
    return dict((m, _align_profile_set(profiles[m], slabs[m], align))
                for m in profiles.keys())

def show_errors(errors, contours=_CONTOURS, npoints=200,
                align='auto', plots=1, save=None):
    """
    Plot the aligned profiles and the distribution of the residuals for
    profiles and residuals returned from calc_errors.

    *contours* can be a list of percentiles or [].  If percentiles
    are given, then show uncertainty using a contour plot with the
    given levels, otherwise just overplot sample lines.
    *contours* defaults to [68, 95, 100].

    *npoints* is the number of points to use when generating the
    profile contour.  Since the z values for the various lines do not
    correspond, the contour generator interpolates the entire profile
    range with linear spacing using this number of points.

    *align* is the interface number plus fractional distance within
    the layer following the interface.  For example, use 0 for the
    substrate interface, use -1 for the surface interface, or use 2.5
    for the center of the second slab above the substrate.

    *plots* is the number of plots to use (1, 2, or 'n').

    *save* is the basename of the plot to save.  This should usually
    be "<store>/<model>".  The program will add '-err#.png' where '#'
    is the number of the plot.
    """
    import matplotlib.pyplot as plt

    if plots == 0: # Don't create plots, just save the data
        _save_profile_data(errors, contours=contours, npoints=npoints,
                           align=align, save=save)
        _save_residual_data(errors, contours=contours, save=save)
    elif plots == 1: # Subplots for profiles/residuals
        plt.subplot(211)
        show_profiles(errors, contours=contours, npoints=npoints, align=align)
        plt.subplot(212)
        show_residuals(errors, contours=contours)
        if save:
            plt.savefig(save+"-err.png")
    elif plots == 2:  # Separate plots for profiles/residuals
        show_profiles(errors, contours=contours, npoints=npoints, align=align)
        if save:
            plt.savefig(save+"-err1.png")
        plt.figure()
        show_residuals(errors, contours=contours)
        if save:
            plt.savefig(save+"-err2.png")
    else: # Multiple plots
        profiles, slabs, Q, residuals = errors
        fignum = 1
        for m in profiles.keys():
            plt.figure()
            show_profiles(errors=({m:profiles[m]}, {m:slabs[m]}, None, None),
                          contours=contours, npoints=npoints, align=align)
            if save:
                plt.savefig(save+"-err%d.png"%fignum)
            fignum += 1
        for m in residuals.keys():
            plt.figure()
            show_residuals(errors=(None, None, {m:Q[m]}, {m:residuals[m]}),
                           contours=contours)
            if save:
                plt.savefig(save+"-err%d.png"%fignum)
            fignum += 1

def show_profiles(errors, align, contours, npoints):
    profiles, slabs, _, _ = errors
    if align is not None:
        profiles = align_profiles(profiles, slabs, align)

    if contours:
        _profiles_contour(profiles, contours, npoints)
    else:
        _profiles_overplot(profiles)


def show_residuals(errors, contours):
    _, _, Q, residuals = errors

    if False and contours:
        _residuals_contour(Q, residuals, contours=contours)
    else:
        _residuals_overplot(Q, residuals)


def _save_profile_data(errors, align, contours, npoints, save):
    profiles, slabs, _, _ = errors
    if align is not None:
        profiles = align_profiles(profiles, slabs, align)
    k = 1
    for title, p in sorted((m.name, p) for m, p in profiles.items()):
        # Find limits of all profiles
        z = np.hstack([line[0] for line in p])
        zp = np.linspace(np.min(z), np.max(z), npoints)
        # Interpolate rho on common z
        rho = np.vstack([np.interp(zp, L[0], L[1]) for L in p])
        q, qval = form_quantiles(rho, contours)
        data = np.vstack((zp, rho[0], np.reshape(qval, (-1, qval.shape[2]))))
        columns = ["z", "best"] + list("%g%%"%v for v in 100*q.flatten())
        _write_file(save+"_rho_contour%d.dat"%k, data, title, columns)
        if len(p[0]) > 3:
            rhoM = np.vstack([np.interp(zp, L[0], L[3]) for L in p])
            q, qval = form_quantiles(rhoM, contours)
            data = np.vstack((zp, rhoM[0], np.reshape(qval, (-1, qval.shape[2]))))
            _write_file(save+"_rhoM_contour%d.dat"%k, data, title, columns)
        k += 1

def _save_residual_data(errors, contours, save):
    _, _, Q, residuals = errors
    k = 1
    for title, x, r in sorted((m.name, Q[m], v) for m, v in residuals.items()):
        q, qval = form_quantiles(r.T, contours)
        # TODO: should have columns for R, dR as well.
        data = np.vstack((x, r[:, 0], np.reshape(qval, (-1, qval.shape[2]))))
        columns = ["q", "best"] + list("%g%%"%v for v in 100*q.flatten())
        _write_file(save+"_resid_contour%d.dat"%k, data, title, columns)
        k += 1

def _write_file(path, data, title, columns):
    with open(path, "wb") as fid:
        fid.write(asbytes("# "+title+"\n"))
        fid.write(asbytes("# "+" ".join(columns)+"\n"))
        np.savetxt(fid, data.T)


# ===== Plotting functions =====

def _profiles_labels(magnetic):
    import matplotlib.pyplot as plt
    plt.xlabel('z (A)')
    if magnetic:
        plt.ylabel('rho (1/A^2)')
    else:
        plt.ylabel('rho, rhoM (1/A^2)')

def _profiles_overplot(profiles):
    import matplotlib.pyplot as plt

    alpha = 0.1
    any_magnetic = False
    for p in profiles.values():
        if len(p[0]) == 3:
            rho_color = next_color()
            for z, rho, _ in p[1:]:
                plt.plot(z, rho, '-', color=rho_color, alpha=alpha)
            # Plot best
            z, rho, _ = p[0]
            plt.plot(z, rho, '-', color=dhsv(rho_color, dv=-0.2))
        else:
            any_magnetic = True
            rho_color = next_color()
            rhoM_color = next_color()
            for z, rho, _, rhoM, _ in p[1:]:
                plt.plot(z, rho, '-', color=rho_color, alpha=alpha)
                plt.plot(z, rhoM, '-', color=rhoM_color, alpha=alpha)
            # Plot best
            z, rho, _, rhoM, _ = p[0]
            plt.plot(z, rho, '-', color=dhsv(rho_color, dv=-0.2))
            plt.plot(z, rhoM, '-', color=dhsv(rhoM_color, dv=-0.2))
    _profiles_labels(any_magnetic)

def _profiles_contour(profiles, contours=_CONTOURS, npoints=200):
    import matplotlib.pyplot as plt

    any_magnetic = False
    for p in profiles.values():
        # Find limits of all profiles
        z = np.hstack([line[0] for line in p])
        zp = np.linspace(np.min(z), np.max(z), npoints)
        if len(p[0]) == 3:
            rho_color = next_color()
            # Interpolate rho on common z
            rho = np.vstack([np.interp(zp, L[0], L[1]) for L in p])
            # Plot the quantiles
            plot_quantiles(zp, rho, contours, rho_color)
            # Plot the best
            plt.plot(zp, rho[0], '-', color=dhsv(rho_color, dv=-0.2))
        else:
            any_magnetic = True
            rho_color = next_color()
            rhoM_color = next_color()
            # Interpolate rho, rhoM on common z
            rho = np.vstack([np.interp(zp, L[0], L[1]) for L in p])
            rhoM = np.vstack([np.interp(zp, L[0], L[3]) for L in p])
            # Plot the quantiles
            plot_quantiles(zp, rho, contours, rho_color)
            plot_quantiles(zp, rhoM, contours, rhoM_color)
            # Plot the best
            plt.plot(zp, rho[0], '-', color=dhsv(rho_color, dv=-0.2))
            plt.plot(zp, rhoM[0], '-', color=dhsv(rhoM_color, dv=-0.2))
    _profiles_labels(any_magnetic)

def _residuals_labels():
    import matplotlib.pyplot as plt

    plt.xlabel('Q (1/A)')
    plt.ylabel('Residuals')

def _residuals_overplot(Q, residuals):
    import matplotlib.pyplot as plt
    alpha = 0.4
    shift = 0
    for m, r in residuals.items():
        color = next_color()
        plt.plot(Q[m], shift+r[:, 1:], '.', markersize=1,
                 color=color, alpha=alpha)
        plt.plot(Q[m], shift+r[:, 0], '.', markersize=1,
                 color=dhsv(color, dv=-0.2)) # best
        shift += 5
    _residuals_labels()

def _residuals_contour(Q, residuals, contours=_CONTOURS):
    import matplotlib.pyplot as plt
    shift = 0
    for m, r in residuals.items():
        color = next_color()
        plot_quantiles(Q[m], shift+r.T, contours, color)
        plt.plot(Q[m], shift+r[:, 0], '.', markersize=1,
                 color=dhsv(color, dv=-0.2)) # best
        shift += 5
    _residuals_labels()

# ==== Helper functions =====

def _align_profile_set(profiles, slabs, align):
    """
    Align all profiles to the first profile.
    """
    p1 = profiles[0]
    t1_offset = _find_offset(slabs[0], align) if align != 'auto' else None
    offsets = [0]
    for p2, t2 in zip(profiles[1:], slabs[1:]):
        offsets.append(_align_profile_pair(p1[0], p1[1], t1_offset,
                                           p2[0], p2[1], t2,
                                           align))
    profiles = [tuple([p[0]+offset]+list(p[1:]))
                for offset, p in zip(offsets, profiles)]
    return profiles

def _align_profile_pair(z1, r1, t1_offset, z2, r2, t2, align):
    """
    Use autocorrelation to align r1 and r2.
    """
    if align == 'auto':
        # Assume z1, z2 have the same step size
        n2 = len(r2)
        idx = np.argmax(np.correlate(r1, r2, 'full'))
        if idx < n2:
            offset = z2[(n2-1)-idx] - z1[0]
        else:
            offset = z2[0] - z1[idx-(n2-1)]
        return -offset
    else:
        return -(_find_offset(t2, align) - t1_offset)

def _find_offset(v, align):
    """
    Find the offset of k.p, where k is the interface number and p is the
    distance into that interface.

    This may even work for interfaces defined from the left, such as
    -1.5 to specify the middle of the final layer.
    """
    idx = int(align)
    offset = np.sum(v[:idx]) + (align-idx)*v[idx]
    #print offset, idx, v[:idx], align
    return offset



import wx

from numpy import inf

from bumps.gui.input_list import InputListPanel

from refl1d.names import Slab, FreeLayer, FreeInterface, PolymerBrush
from refl1d.material import SLD

# name, type, description, *args
# type: 'string'
# type: 'material' (see below)
# type: 'parameter' with units and limits (see below)
# type: '[parameter]' is a vector of parameters with common properties

# 'material' is a scatterer, which is one of:
#
#   SLD, with scattering length density rho + irho
#   Material, with compound and natural density, which can be fit
#      by density, natural density, relative density or cell volume.
#      The compound may be a chemical formula, or it may be multiple
#      formulas, with a count for each.
#   Mixture, which is a set of scatterers with a base making up the
#      bulk and % making up the rest.  The % is by mass or by volume.
#      If by mass, each material needs a density.  The scatterers can
#      be SLDs, materials or other mixtures.
#
# The material name defaults to the chemical formula, if available.
#
# Users must be able to select a material from another layer, with
# fittable parameters automatically tied to that other layer, and a
# visual indication that this is so.
#
# A database of scatterers used in various experiments would be useful.
#
# The periodictable class can be used to lookup expected SLD values
# as in the SANS SLD calculator.
#
# 'parameter' has name, value, and can be fixed, fitted or computed.
#
# Fitted parameters have a range, which can be value pm v, value pm v%,
# value pm (-v1,v2), value pm (-v1,+v2)%, or simply [v1,v2], where value
# is the current parameter value.  Maybe we can do this in the input
# box for value.
#
# Computed values need a formula, which can be any math expression,
# with a popup box to select parameter Pn from a table of Pn:name:path.
#
# Parameters may be copied directly from other fields with the same units,
# presumably using a combobox to select the parameter.
#
# Some of these operations will require more information than just the
# current stack.

# These belong with the layer definition, so that users can add their own
# layer types more easily.  I will use monkey patching until the details
# of the interface are worked out.
Slab.name = 'slab'
Slab.description = 'layer with uniform density and roughness on top'
Slab.fields = (
    ('name','string','slab name (defaults to material name)'),
    ('material','material','slab material'),
    ('thickness','parameter','slab thickness','A',(0,inf)),
    ('interface','parameter',
     'rms roughness between this slab and the next','A',(0,inf)),
)

PolymerBrush.name = 'brush'
PolymerBrush.description = 'polymer brush in a solvent'
PolymerBrush.fields = (
    ('name','string','brush name (defaults to polymer name)'),
    ('polymer','material','polymer composition'),
    ('solvent','material','solvent composition'),
    ('thickness','parameter','brush+solvent thickness','A',(0,inf)),
    ('interface','parameter',
     'rms roughness between the solvent and the next layer','A',(0,inf)),
    ('base_vf','parameter',
     'volume fraction of the polymer brush at the base','%',(0,100)),
    ('base','parameter','thickness of the base region','A',(0,inf)),
    ('length','parameter','thickness of the thinning region','A',(0,inf)),
    ('power','parameter','rate of brush thinning','',(-inf,inf)),
    ('sigma','parameter','rms roughness within the brush','A',(0,inf)),
)

FreeLayer.name = 'freeform'
FreeLayer.description = 'freeform layer using monotonic splines for good control'
FreeLayer.fields = (
    ('name','string','layer name'),
    ('thickness','string','total layer thickness','A',(0,inf)),
    # The following need to be links to neighbouring layer parameters; easy
    # enough with slabs, but somewhat messy with something like a tethered
    # polymer.  Perhaps we can define below and above fields for all the
    # layer types which act as scatterers to make this happen, but for now
    # just assume it is attached to a slab.  Insert/Delete are going to
    # have to update
    ('below','material','material below the layer'),
    ('above','material','material above the layer'),
    # Need to indicate that these are three coordinated vectors
    ('z','[parameter]','control point location','A',(0,1)),
    ('rho','[parameter]',
     'scattering length density','1e-6 inv A^2',(-inf,inf)),
    ('irho','[parameter]',
     'complex scattering length density','1e-6 inv A^2',(0,inf)),
)

FreeInterface.name = 'blend'
FreeInterface.description = 'blend between two materials using a monotonic spline'
FreeInterface.fields = (
    ('name','string','layer name'),
    ('below','material','material below the layer'),
    ('above','material','material above the layer'),
    ('thickness','parameter','total layer thickness','A',(0,inf)),
    ('interface','parameter',
     'rms roughness between "above" and the next layer', 'A', (0,inf)),
    # Need to indicate that these are two coordinated vectors
    ('dz','[parameter]','relative segment size','',(0,inf)),
    ('dp','[parameter]','relative step height','',(0,inf)),
)

LAYERS = (Slab, FreeInterface, FreeLayer, PolymerBrush)


class LayerEditorDialog(wx.Dialog):
    """
    Select layer type and edit layer parameters interactively.
    """
    def __init__(self,
                 parent = None,
                 id = wx.ID_ANY, #@ReservedAssignment
                 title = "Layer Editor",
                 pos = wx.DefaultPosition,
                 size = (-1, 400),
                 style = wx.DEFAULT_FRAME_STYLE, #|wx.THICK_FRAME|wx.RESIZE_BORDER,
                 name = "",
                 stack = None,
                 layer_num = 0):
        wx.Dialog.__init__(self, parent, id, title, pos, size, style, name)
        if parent is not None: self.SetFont(parent.GetFont())

        self.next_button = wx.Button(self, wx.ID_ANY, "Next", style=wx.BU_EXACTFIT)
        self.prev_button = wx.Button(self, wx.ID_ANY, "Prev", style=wx.BU_EXACTFIT)
        self.ins_button = wx.Button(self, wx.ID_ANY, "Insert", style=wx.BU_EXACTFIT)
        self.del_button = wx.Button(self, wx.ID_ANY, "Delete", style=wx.BU_EXACTFIT)

        self.layer_id = wx.StaticText(self, size=(-1,-1))
        self.selector = wx.Choice(self)
        self.fields = InputListPanel(self)

        # Generic bindings
        self.Bind(wx.EVT_BUTTON, self.OnNext, self.next_button)
        self.Bind(wx.EVT_BUTTON, self.OnPrev, self.prev_button)
        self.Bind(wx.EVT_BUTTON, self.OnInsert, self.ins_button)
        self.Bind(wx.EVT_BUTTON, self.OnDelete, self.del_button)
        self.Bind(wx.EVT_TEXT, self.OnValidate)
        #self.Bind(wx.EVT_COMBO_BOX, self.OnValidate)
        self.selector.Bind(wx.EVT_CHOICE, self.OnLayerType)

        # Layout
        outer = wx.BoxSizer(wx.VERTICAL)

        bbox = wx.GridSizer(rows=1, cols=4, vgap=0, hgap=5)
        bbox.Add(self.next_button, 1, wx.EXPAND)
        bbox.Add(self.prev_button, 1, wx.EXPAND)
        bbox.Add(self.ins_button, 1, wx.EXPAND)
        bbox.Add(self.del_button, 1, wx.EXPAND)
        outer.Add(bbox, 0, wx.ALIGN_LEFT|wx.LEFT|wx.RIGHT|wx.TOP, border=10)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        hbox.Add(self.layer_id, 0,
                 wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL|wx.RIGHT, border=5)
        hbox.Add(self.selector,0, wx.ALIGN_LEFT|wx.ALIGN_CENTER_VERTICAL)
        hbox.Add((0,0), 1)
        outer.Add(hbox, 0, wx.EXPAND|wx.LEFT|wx.RIGHT|wx.TOP, border=10)
        outer.Add(self.fields, 1, wx.EXPAND|wx.LEFT|wx.RIGHT|wx.TOP, border=10)
        outer.Add((-1,10))

        dx,dy = outer.GetMinSize()
        outer.SetMinSize((dx,dy+100))
        outer.Fit(self)
        self.SetSizer(outer)

        # Set initial layer
        self.layer = None
        self.layer_num = layer_num
        self.set_stack(stack)

    def OnNext(self, event):
        self.set_layer(self.layer_num+1)
    def OnPrev(self, event):
        self.set_layer(self.layer_num-1)
    def OnDelete(self, event):
        del self.stack[self.layer_num]
        self.set_layer(self.layer_num)
    def OnInsert(self, event):
        slab = Slab(SLD(rho=1.0,name="layer"), thickness=100, interface=5)
        self.stack.insert(self.layer_num, slab)
        self.set_layer(self.layer_num)

    def OnLayerType(self, event):
        pass

    def set_stack(self, stack):
        if stack is None: return
        self.stack = stack
        self.set_layer(self.layer_num)

    def set_layer(self, layer_num):
        # Make sure the layer is valid
        N = len(self.stack)
        if layer_num < 0: layer_num = 0
        elif layer_num >= N: layer_num = N-1
        self.layer_num = layer_num

        # Enable navigation and control buttons
        self.next_button.Enable(layer_num < N-1)
        self.prev_button.Enable(layer_num > 0)
        self.del_button.Enable(N > 1)
        self.layer_id.SetLabel("Layer %d:"%layer_num)

        # Null case: no stack
        if N == 0: return # Should never happen...

        # Update the layer contents
        if self.layer != self.stack[layer_num]:
            self.layer = self.stack[layer_num]
            self.draw_layer()

    def draw_layer(self):
        return
        if self.layer is None: return
        self.name_label = wx.StaticText(self, label="Name")
        self.name = wx.TextCtrl(self, wx.ID_ANY, size=(250,-1))

        self.thickness_label = wx.StaticText(self, label="Thickness")
        self.thickness = wx.TextCtrl(self, wx.ID_ANY, size=(100,-1))
        self.interface_label = wx.StaticText(self, label="Interface")
        self.interface = wx.TextCtrl(self, wx.ID_ANY, size=(100,-1))

        self.layer_name.SetValue(self.layer.name)
        self.thickness.SetValue("%g"%self.layer.thickness.value)
        self.interface.SetValue("%g"%self.layer.interface.value)
        #self.selector.SetValue(self.layer.__class__.__name__)

    def OnValidate(self, event):
        ctrl = event.GetEventObject()
        if ctrl:
            validator = ctrl.GetValidator()
            if validator:
                validator.Validate(ctrl)
        event.Skip()

def main(stack):
    class App(wx.App):
        def OnInit(self):
            dia = LayerEditorDialog(None, -1, "Mock-up")
            dia.set_stack(stack)
            dia.ShowModal()
            dia.Destroy()
            return True

    app = App(0)
    app.MainLoop()

if __name__ == "__main__":
    from refl1d.names import silicon, air
    stack = silicon | air
    main(stack)




def _build_from_dict(experiment):
    pass
    # remap
def load_json(path):
    import json
    with open(path, 'r') as fid:
        data = fid.read()
    experiment = json.loads(data)
    return _build_from_dict(experiment)



#!/usr/bin/env python
import sys
import os

if len(sys.argv) == 1:
    sys.argv.append('install')

# Use our own nose-based test harness
if sys.argv[1] == 'test':
    from subprocess import call
    sys.exit(call([sys.executable, 'test.py']+sys.argv[2:]))

#sys.dont_write_bytecode = True

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

packages = ['refl1d', 'refl1d.view']

version = None
for line in open(os.path.join("refl1d", "__init__.py")):
    if "__version__" in line:
        version = line.split('"')[1]
#print("Version: "+version)

extra_compile_args =  {'msvc': ['/EHsc']}
extra_link_args =  {}

class build_ext_subclass(build_ext):
    def build_extensions(self):
        c = self.compiler.compiler_type
        if c in extra_compile_args:
           for e in self.extensions:
               e.extra_compile_args = extra_compile_args[c]
        if c in extra_link_args:
            for e in self.extensions:
                e.extra_link_args = extra_link_args[c]
        build_ext.build_extensions(self)

# reflmodule extension
def reflmodule_config():
    if sys.platform == "darwin":
        # Python is not finding C++ headers on Mac unless the
        # minimum OSX version is bumped from the default 10.6 up
        # to 10.10.  Don't know if this is because of the mac
        # setup (older development libraries not installed) or
        # because of the anaconda build (targetted to 10.6) or
        # some combination.  Override by setting the deployment
        # target on the command line.  Curiously, Xcode can
        # target c++ code to 10.7 on the same machine.
        #os.environ.setdefault('MACOSX_DEPLOYMENT_TARGET', '10.10')
        os.environ.setdefault('MACOSX_DEPLOYMENT_TARGET', '10.13')

    S = ("reflmodule.cc", "methods.cc",
         "reflectivity.cc", "magnetic.cc",
         "contract_profile.cc",
         "convolve.c", "convolve_sampled.c",
        )

    Sdeps = ("erf.c", "methods.h", "rebin.h", "rebin2D.h", "reflcalc.h")
    sources = [os.path.join('refl1d', 'lib', f) for f in S]
    depends = [os.path.join('refl1d', 'lib', f) for f in Sdeps]
    return Extension('refl1d.reflmodule',
                     sources=sources,
                     depends=depends,
                     language="c++",
                     #py_limited_api="cp32",
                     )

# SCF extension
def SCFmodule_config():
    from numpy import get_include
    return Extension('refl1d.calc_g_zs_cex',
                     sources=[os.path.join('refl1d', 'lib', 'calc_g_zs_cex.c')],
                     include_dirs=[get_include()],
                     #language="c++",
                     #py_limited_api="cp32",
                     )

#TODO: write a proper dependency checker for packages which cannot be
# installed by easy_install
#dependency_check('numpy>=1.0', 'scipy>=0.6', 'matplotlib>=1.0', 'wx>=2.8.9')

#sys.dont_write_bytecode = False
dist = setup(
    name='refl1d',
    version=version,
    author='Paul Kienzle',
    author_email='pkienzle@nist.gov',
    url='http://www.reflectometry.org/danse/model1d.html',
    description='1-D reflectometry modeling',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: Public Domain',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        ],
    packages=packages,
    #package_data=gui_resources.package_data(),
    scripts=['bin/refl1d_cli.py', 'bin/refl1d_gui.py'],
    entry_points={
        'console_scripts': ['refl1d=refl1d.main:cli'],
        'gui_scripts': ['refl1d_gui=refl1d.main:gui']
    },
    ext_modules=[reflmodule_config(), SCFmodule_config()],
    install_requires=['bumps>=0.7.11', 'numpy', 'scipy', 'matplotlib', 'wxpython', 'periodictable'],
    cmdclass={'build_ext': build_ext_subclass},
    )

# End of file



#!/usr/bin/env python

# Copyright (C) 2006-2011, University of Maryland
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/ or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Author: James Krycka, Paul Kienzle

"""
This script builds the Refl1D application and documentation from source and
runs unit tests and doc tests.  It supports building on Windows and Linux.

Usually, you downloaded this script into a top-level directory (the root)
and run it from there which downloads the files from the application
repository into a subdirectory (the package directory).  For example if
test1 is the root directory, we might have:
  E:/work/test1/master_builder.py
               /refl1d/master_builder.py
               /refl1d/...

Alternatively, you can download the whole application repository and run
this script from the application's package directory where it is stored.
The script determines whether it is executing from the root or the package
directory and makes the necessary adjustments.  In this case, the root
directory is defined as one-level-up and the repository is not downloaded
(as it is assumed to be fully present).  In the example below test1 is the
implicit root (i.e. top-level) directory.
  E:/work/test1/refl1d/master_builder.py
               /refl1d/...

Getting "git pull" support on Windows requires some work.  You first need to
install Git Bash from msysgit (github.com has a link).  To build, start
gitbash, and set up your ssh keys using::

    $ eval `ssh-agent -s`
    $ ssh-add

If you want to use PuTTY instead, you must convert your private key to a
putty key and drop it on your desktop.  Click the private key icon to start
pagent with your key.  Uncomment os.environ['GIT_SSH'] below, and set it to
the location of plink.exe on your system.
"""
from __future__ import print_function

import os
import sys
import shutil
import subprocess
from distutils.util import get_platform

# python 3 uses input rather than raw_input
try:
    input = raw_input
except Exception:
    pass

GIT = "git"
MAKE = "make"

# Windows commands to run utilities
PYTHON = sys.executable
if os.name == "nt":
    SYSBIN = r"C:\Program Files (x86)"
    if not os.path.exists(SYSBIN):
        SYSBIN = r"C:\Program Files"
    GIT = SYSBIN+r"\Git\mingw64\bin\git.exe"
    INNO = SYSBIN+r"\Inno Setup 5\ISCC.exe"  # command line operation

    if not os.path.exists(GIT):
        print("missing git: "+GIT+" --- source will not be updated",
              file=sys.stderr)
    if not os.path.exists(INNO):
        print("missing inno setup: "+INNO+" --- installer will not be built",
              file=sys.stderr)

    # Put PYTHON in the environment and add the python directory and its
    # corresponding script directory (for nose, sphinx, pip, etc) to the path.
    PYTHONDIR = os.path.dirname(os.path.abspath(PYTHON))
    SCRIPTDIR = os.path.join(PYTHONDIR, 'Scripts')
    #LATEXDIR = 'C:/Program Files/MikTeX 2.9/miktex/bin/x64'
    LATEXDIR = 'z:/miktex/miktex/bin'  # Portable miktex distribution
    os.environ['PATH'] = ";".join((PYTHONDIR, SCRIPTDIR, LATEXDIR, os.environ['PATH']))
    os.environ['PYTHON'] = "/".join(PYTHON.split("\\"))
    #os.environ['GIT_SSH'] = r"C:\Program Files (x86)\PuTTY\plink.exe"
    MAKE = r"C:\mingw\bin\mingw32-make"
    #MAKE = r"make.bat"
else:
    # Support for wx in virtualenv on mac
    if hasattr(sys, "real_prefix"):
        PYTHON = os.path.join(sys.real_prefix, 'bin', 'python')
        os.environ['PYTHON'] = PYTHON
        os.environ['PYTHONHOME'] = sys.prefix

# Name of the package
PKG_NAME = "refl1d"
# Name of the application we're building
APP_NAME = "Refl1D"


# Required versions of Python packages and utilities to build the application.
MIN_PYTHON = "2.6"
MAX_PYTHON = "3.0"
MIN_MATPLOTLIB = "1.0.0"
MIN_NUMPY = "1.3.0"
MIN_SCIPY = "0.7.0"
MIN_WXPYTHON = "2.8.10.1"
MIN_SETUPTOOLS = "0.6c9"
MIN_GCC = "3.4.4"
MIN_PYPARSING = "1.5.2"
MIN_PERIODICTABLE = "1.3"
# Required versions of Python packages to run tests.
MIN_NOSE = "0.11"
# Required versions of Python packages and utilities to build documentation.
MIN_SPHINX = "1.0.3"
MIN_DOCUTILS = "0.5"
MIN_PYGMENTS = "1.0"
MIN_JINJA2 = "2.5.2"
#MIN_MATHJAX = "1.0.1"
# Required versions of Python packages and utilities to build Windows frozen
# image and Windows installer.
MIN_PY2EXE = "0.6.9"
MIN_INNO = "5.3.10"

# Determine the full directory paths of the top-level, source, and installation
# directories based on the directory where the script is running.  Here the
# top-level directory refers to the parent directory of the package.
RUN_DIR = os.path.dirname(os.path.abspath(sys.argv[0]))
head, tail = os.path.split(RUN_DIR)
if tail == PKG_NAME:
    TOP_DIR = head
else:
    TOP_DIR = RUN_DIR

# Set the python path
PLATFORM = '.%s-%s'%(get_platform(), sys.version[:3])
BUMPS_DIR = os.path.join(TOP_DIR, "bumps")
REFL_DIR = os.path.join(TOP_DIR, "refl1d")
PERIODICTABLE_DIR = os.path.join(TOP_DIR, "periodictable")
BUMPS_LIB = os.path.join(BUMPS_DIR, 'build', 'lib'+PLATFORM)
REFL_LIB = os.path.join(REFL_DIR, 'build', 'lib'+PLATFORM)
paths = [BUMPS_LIB, REFL_LIB, PERIODICTABLE_DIR]
sys.path = paths + sys.path
os.environ['PYTHONPATH'] = os.pathsep.join(paths)

BUMPS_NEW = '"%s" clone https://github.com/bumps/bumps.git'%GIT
REFL_NEW = '"%s" clone https://github.com/reflectometry/refl1d.git'%GIT
PERIODICTABLE_NEW = '"%s" clone https://github.com/pkienzle/periodictable.git'%GIT
REPO_UPDATE = '"%s" pull'%GIT

# Note: can't put target files in the dist directory since all files in dist
# are being collected and stored in the bundle.
INSTALLER_DIR = REFL_DIR


def get_version():
    "Determine package version"
    global PKG_VERSION
    # Get the version string of the application for use later.
    # This has to be done after we have checked out the repository.
    for line in open(os.path.join(REFL_DIR, PKG_NAME, '__init__.py')).readlines():
        if line.startswith('__version__'):
            PKG_VERSION = line.split('=')[1].strip()[1:-1]
            break
    else:
        raise RuntimeError("Could not find package version")

#==============================================================================

def checkout_code():
    "download or update source code"
    print("Checking out application code from the repository ...\n")

    if not os.path.exists(GIT):
        print("missing git --- source not checked out")
        return

    if RUN_DIR == TOP_DIR:
        exec_cmd(PERIODICTABLE_NEW, cwd=TOP_DIR)
        exec_cmd(BUMPS_NEW, cwd=TOP_DIR)
        exec_cmd(REFL_NEW, cwd=TOP_DIR)
    else:
        exec_cmd(REPO_UPDATE, cwd=PERIODICTABLE_DIR)
        exec_cmd(REPO_UPDATE, cwd=BUMPS_DIR)
        exec_cmd(REPO_UPDATE, cwd=REFL_DIR)

    get_version()  # reset version number in case it was updated remotely


def create_archive(version=None):
    "create source distribution"
    if version is None:
        version = PKG_VERSION

    # Create zip and tar archives of the source code and a manifest file
    # containing the names of all files.
    print("Creating an archive of the source code ...\n")

    try:
        # Create zip and tar archives in the dist subdirectory.
        exec_cmd("%s setup.py sdist --formats=zip,gztar --dist-dir=%s"
                 %(PYTHON, INSTALLER_DIR), cwd=REFL_DIR)
        exec_cmd("%s setup.py sdist --formats=zip,gztar --dist-dir=%s"
                 %(PYTHON, INSTALLER_DIR), cwd=BUMPS_DIR)
    except Exception:
        print("*** Failed to create source archive ***")


def build_package():
    "build and install the package"
    print("Building the %s package ...\n" %(PKG_NAME,))

    exec_cmd("%s setup.py build" %(PYTHON,), cwd=BUMPS_DIR)
    exec_cmd("%s setup.py build" %(PYTHON,), cwd=REFL_DIR)


def build_documentation():
    "build the documentation"
    print("Running the Sphinx utility to build documentation ...\n")
    doc_dir = os.path.join(REFL_DIR, "doc")
    html_dir = os.path.join(doc_dir, "_build", "html")
    latex_dir = os.path.join(doc_dir, "_build", "latex")
    pdf = os.path.join(latex_dir, APP_NAME+".pdf")

    # Delete any left over files from a previous build.
    # Create documentation in HTML and PDF format.
    sphinx_cmd = '"%s" -m sphinx.__init__ -b %%s -d _build/doctrees -D latex_paper_size=letter . %%s'%PYTHON
    exec_cmd(sphinx_cmd%("html", html_dir), cwd=doc_dir)
    if True:  # build pdf as well
        exec_cmd(sphinx_cmd%("latex", latex_dir), cwd=doc_dir)
        exec_cmd("pdflatex %s.tex"%APP_NAME, cwd=latex_dir)
        exec_cmd("pdflatex %s.tex"%APP_NAME, cwd=latex_dir)
        exec_cmd("pdflatex %s.tex"%APP_NAME, cwd=latex_dir)
        exec_cmd("makeindex -s python.ist %s.idx"%APP_NAME, cwd=latex_dir)
        exec_cmd("pdflatex %s.tex"%APP_NAME, cwd=latex_dir)
        exec_cmd("pdflatex %s.tex"%APP_NAME, cwd=latex_dir)
        # Copy PDF to the html directory where the html can find it.
        if os.path.isfile(pdf):
            shutil.copy(pdf, html_dir)


def create_binary():
    "create the standalone executable (windows or mac, depending)"
    if sys.platform == 'darwin':
        print("Using py2app to create a Mac OS X app...\n")
        exec_cmd("%s setup_py2app.py"%PYTHON, cwd=REFL_DIR)

    elif os.name == 'nt':
        print("Using py2exe to create a Win32 executable ...\n")
        exec_cmd("%s setup_py2exe.py"%PYTHON, cwd=REFL_DIR)

    else:
        print("No binary build for %s %s\n" % (sys.platform, os.name))


def create_installer(version=None):
    if not version:
        version = PKG_VERSION
    if not os.path.exists(INSTALLER_DIR):
        os.mkdir(INSTALLER_DIR)
    if sys.platform == 'darwin':
        _create_osx_installer(version)
    elif os.name == 'nt':
        _create_windows_installer(version)
    else:
        print("No installer for %r"%sys.platform)

def _create_osx_installer(version):
    "create mac dmg"
    print("Running command to create the dmg")
    exec_cmd([PYTHON, "extra/build_dmg.py", APP_NAME, PKG_VERSION],
             cwd=REFL_DIR)

def _create_windows_installer(version):
    "create the windows installer"
    if not os.path.exists(INNO):
        print("missing INNO --- no installer")
    # Run the Inno Setup Compiler to create a Win32 installer/uninstaller for
    # the application.
    print("Running Inno Setup Compiler to create Win32 "
          "installer/uninstaller ...\n")

    # Run the Inno Setup Compiler to create a Win32 installer/uninstaller.
    # Override the output specification in <PKG_NAME>.iss to put the executable
    # and the manifest file in the top-level directory.
    arch = "x64" if sys.maxsize > 2**32 else "x86"
    exec_cmd([INNO, "/Q", "/O"+INSTALLER_DIR,
              "/DVERSION="+version, "/DARCH="+arch, PKG_NAME+".iss"],
             cwd=REFL_DIR)


def run_tests():
    "run the test suite"
    # Run unittests and doctests using a test script.
    # Running from a test script allows customization of the system path.
    print("Running periodictable tests ...\n")
    exec_cmd("%s test.py"%PYTHON, cwd=PERIODICTABLE_DIR)
    print("Running bumps tests ...\n")
    exec_cmd("%s test.py"%PYTHON, cwd=BUMPS_DIR)
    print("Running refl tests ...\n")
    exec_cmd("%s test.py"%PYTHON, cwd=REFL_DIR)

def check_dependencies():
    "check that required packages are installed"

    import platform
    from pkg_resources import parse_version as PV

    # ------------------------------------------------------
    python_ver = platform.python_version()
    print("Using Python %s"%python_ver)
    print("")
    if PV(python_ver) < PV(MIN_PYTHON) or PV(python_ver) >= PV(MAX_PYTHON):
        print("ERROR - build requires Python >= %s, but < %s"
              %(MIN_PYTHON, MAX_PYTHON))
        sys.exit()

    req_pkg = {}

    # ------------------------------------------------------
    try:
        from matplotlib import __version__ as mpl_ver
    except Exception:
        mpl_ver = "0"
    finally:
        req_pkg["matplotlib"] = (mpl_ver, MIN_MATPLOTLIB)

    # ------------------------------------------------------
    try:
        from numpy import __version__ as numpy_ver
    except Exception:
        numpy_ver = "0"
    finally:
        req_pkg["numpy"] = (numpy_ver, MIN_NUMPY)

    # ------------------------------------------------------
    try:
        from scipy import __version__ as scipy_ver
    except Exception:
        scipy_ver = "0"
    finally:
        req_pkg["scipy"] = (scipy_ver, MIN_SCIPY)

    # ------------------------------------------------------
    try:
        from wx import __version__ as wx_ver
    except Exception:
        wx_ver = "0"
    finally:
        req_pkg["wxpython"] = (wx_ver, MIN_WXPYTHON)

    # ------------------------------------------------------
    try:
        from setuptools import __version__ as setup_ver
    except Exception:
        setup_ver = "0"
    finally:
        req_pkg["setuptools"] = (setup_ver, MIN_SETUPTOOLS)

    # ------------------------------------------------------
    try:
        shell = os.name != 'nt'
        p = subprocess.Popen("gcc -dumpversion", stdout=subprocess.PIPE,
                             shell=shell)
        gcc_ver = p.stdout.read().strip()
    except Exception:
        gcc_ver = "0"
    finally:
        req_pkg["gcc"] = (gcc_ver, MIN_GCC)

    # ------------------------------------------------------
    try:
        from pyparsing import __version__ as parse_ver
    except Exception:
        parse_ver = "0"
    finally:
        req_pkg["pyparsing"] = (parse_ver, MIN_PYPARSING)

    # ------------------------------------------------------
    try:
        from periodictable import __version__ as ptab_ver
    except Exception:
        ptab_ver = "0"
    finally:
        req_pkg["periodictable"] = (ptab_ver, MIN_PERIODICTABLE)

    # ------------------------------------------------------
    try:
        from nose import __version__ as nose_ver
    except Exception:
        nose_ver = "0"
    finally:
        req_pkg["nose"] = (nose_ver, MIN_NOSE)

    # ------------------------------------------------------
    try:
        from sphinx import __version__ as sphinx_ver
    except Exception:
        sphinx_ver = "0"
    finally:
        req_pkg["sphinx"] = (sphinx_ver, MIN_SPHINX)

    # ------------------------------------------------------
    try:
        from docutils import __version__ as docutils_ver
    except Exception:
        docutils_ver = "0"
    finally:
        req_pkg["docutils"] = (docutils_ver, MIN_DOCUTILS)

    # ------------------------------------------------------
    try:
        from pygments import __version__ as pygments_ver
    except Exception:
        pygments_ver = "0"
    finally:
        req_pkg["pygments"] = (pygments_ver, MIN_PYGMENTS)

    # ------------------------------------------------------
    try:
        from jinja2 import __version__ as jinja2_ver
    except Exception:
        jinja2_ver = "0"
    finally:
        req_pkg["jinja2"] = (jinja2_ver, MIN_JINJA2)

    # ------------------------------------------------------
    if os.name == 'nt':
        try:
            from py2exe import __version__ as py2exe_ver
        except Exception:
            py2exe_ver = "0"
        finally:
            req_pkg["py2exe"] = (py2exe_ver, MIN_PY2EXE)

        if os.path.isfile(INNO):
            req_pkg["Inno Setup Compiler"] = ("?", MIN_INNO)
        else:
            req_pkg["Inno Setup Compiler"] = ("0", MIN_INNO)

    # ------------------------------------------------------
    error = False
    for key in req_pkg:
        if req_pkg[key][0] == "0":
            print("====> %s not found; version %s or later is required - ERROR"
                  % (key, req_pkg[key][1]))
            error = True
        elif req_pkg[key][0] == "?":
            print("Found %s" %(key))  # version is unknown
        elif PV(req_pkg[key][0]) >= PV(req_pkg[key][1]):
            print("Found %s %s" %(key, req_pkg[key][0]))
        else:
            print("Found %s %s but minimum tested version is %s - WARNING"
                  % (key, req_pkg[key][0], req_pkg[key][1]))
            error = True

    if error:
        ans = input("\nDo you want to continue (Y|N)? [N]: ")
        if ans.upper() != "Y":
            sys.exit()
    else:
        print("\nSoftware dependencies have been satisfied")


def exec_cmd(command, cwd=None):
    """Runs the specified command in a subprocess."""

    if isinstance(command, list):
        command_str = " ".join('"%s"' % p if ' ' in p else p for p in command)
        shell = os.name == 'nt'
    else:
        command_str = command
        shell = True
    print("%s$ %s" % (os.getcwd(), command_str))
    result = subprocess.call(command, shell=shell, cwd=cwd)
    if result != 0:
        sys.exit(result)

BUILD_POINTS = [
    ('deps', check_dependencies),
    ('update', checkout_code),
    ('build', build_package),
    ('test', run_tests),
    ('docs', build_documentation),  # Needed by windows installer
    ('zip', create_archive),
    ('exe', create_binary),
    ('installer', create_installer),
]

def main():
    points, _ = zip(*BUILD_POINTS)
    start = BUILD_POINTS[0][0]
    only = False
    if len(sys.argv) > 1:
        # Display help if requested.
        if (len(sys.argv) > 1 and sys.argv[1] not in points
                or len(sys.argv) > 2 and sys.argv[2] != 'only'
                or len(sys.argv) > 3):
            print("\nUsage: python master_builder.py [<start>] [only]\n")
            print("Build start points:")
            for point, function in BUILD_POINTS:
                print("  %-10s %s"%(point, function.__doc__))
            print("Add 'only' to the command to only perform a single step")
            sys.exit()
        else:
            start = sys.argv[1]
            only = len(sys.argv) > 2 and sys.argv[2] == 'only'

    get_version()
    print("\nBuilding the %s-%s application from the %s repository ...\n"
          % (APP_NAME, PKG_VERSION, PKG_NAME))
    print("Current working directory  = %s"%RUN_DIR)
    print("Top-level (root) directory = %s"%TOP_DIR)
    print("Package (source) directory = %s"%REFL_DIR)

    started = False
    for point, function in BUILD_POINTS:
        if point == start:
            started = True
        if not started:
            continue
        print("%s %s %s"%("/"*5, point, "/"*25))
        function()
        if only:
            break

if __name__ == "__main__":
    main()



# This program is public domain
# Author: Paul Kienzle
r"""
Interfacial roughness

Interfacial roughness is defined by a probability distribution P and the
step function H, with the convolution (P*H)(z) defining the probability
of finding a part of layer B in A.

Example
-------

Define a silicon wafer with tanh roughness between 0 and 5:

    >>> from refl1d.names import *
    >>> sample = silicon(interface=Tanh(5, "Si:Air")) | air

The above example uses a function interface.  Other interface include:

    Sharp()             Dirac delta interface (cumulative density ~ sign(z))
    Erf(width=sigma)    Gaussian interface (cumulative density ~ erf(z))
    Tanh(width=sigma)   sech**2 interface (cumulative density ~ tanh(z))
    Linear(width=w)     boxcar interface (cumulative density ~ linear(z))

You can plot the available interfaces using :func:`demo`.

As you can see from the above plot, the hyperbolic tangent profile will be
indistinguishable from the error function profile in all practical
circumstances.  Given the computational advantage of the error function
profile, there will be little reason to choose anything else.

Some interfaces can be defined using full width at half maximum (FWHM)
rather than 1-\ $\sigma$::

    Erf.as_fwhm(width=fwhm)  Gaussian interface using FWHM
    Tanh.as_fwhm(width=fwhm) sech**2 interface using FWHM

To see what the effect of choosing width FWHM, use :func:`demo_fwhm`

Defined using FWHM, tanh and erf profiles are significantly different
for the same value of width.  From the 1-\ $\sigma$ plots above, though, we
know that some values of width for tanh will closely match the shape
of the erf sigmoid.  What this means is that you will get different
values for the interface width from *Erf.as_fwhm* and *Tanh.as_fwhm*, but
the generated scattering length density profiles will be indistinguishable.

Extensions
==========

You can define new interface profiles by subclassing from Interface.
You will need to provide the following methods::

    pdf(z)
        returns the probability density function over z for the underlying
        distribution.
    cdf(z)
        returns the cumulative density function, which is the integral
        over the pdf from -inf to z.
    ppf(z)
        returns the percent point function, which is the inverse of
        the cdf.  You can define a profile with n equal sized steps in
        y using y=np.linspace(0, 1, n+2)[1:-1], x=interface.ppf(y)
    parameters()
        returns the set of parameters used to define the interface

See the implementation of :class:`Erf` or :class:`Tanh` for a complete example.
"""
from __future__ import division, print_function

__all__ = ['Interface', 'Sharp', 'Erf', 'Tanh', 'Linear']

import numpy as np
from numpy import tanh, cosh, exp, log, sqrt, pi, inf
from numpy import arccosh as acosh
from numpy import arctanh as atanh
from scipy.special import erf, erfinv

try:
    from bumps.parameter import Parameter
except ImportError:
    print("Could not import Parameter; using trivial implementation")
    class Parameter(object):
        def __init__(self, value, **kw):
            self.value = value
        @classmethod
        def default(cls, value, **kw):
            if isinstance(value, Parameter):
                return value
            else:
                return cls(value)

sech = lambda x: 1/cosh(x)
asech = lambda x: acosh(1/x)

class Interface(object):
    """
    Interfacial mixing function.

    An interface defines the transition from one layer to another in terms
    of the relative proportion of materials on either side of the interface.
    """
    def parameters(self):
        """
        Fittable parameters
        """
        return []

    def cdf(self, z):
        """
        Return the cumulative density function corresponding to the interface.
        """

    def pdf(self, z):
        """
        Return the probability density function corresponding to the interface.
        """

    def ppf(self, z):
        """
        Return the percent point function, which is the inverse of the cdf.
        """

class Sharp(Interface):
    """
    Perfectly sharp interface

    The sharp interface has the form::

        CDF(z) = 1 if z>=0, 0 otherwise
        PDF(z) = inf if z==0, 0 otherwise
        PPF(z) = 0

    This interface has a trivial analytic solution in that it has no
    effect on the optical matrix calculation of the reflectivity.
    """
    def parameters(self):
        return []

    def cdf(self, z):
        return 1*(z >= 0)

    def pdf(self, z):
        return inf*(z == 0)

    def ppf(self, z):
        return 0*z

class Erf(Interface):
    r"""
    Error function profile

    *width* (Parameter: 0 Angstroms)

        1-\ $\sigma$ roughness.  For roughness w measured by the
        full width at half maximum (FWHM), use :meth:`Erf.as_fwhm`\ (w).

    *name* (string: "erf")

    The erf profile has the form:

        CDF(z) = $(1 - \text{erf}(z/(\sqrt{2} \sigma)))$
        PDF(z) = $1/\sqrt{2 \pi \sigma**2} \exp( (z/\sigma)^2/2 )$
        PPF(z) = $\sigma*\sqrt{2}*\text{erf}^{-1}(2*z-1)$

    To convert from a 1-\ $\sigma$ error function to the equivalent FWHM,
    you need to scale the error function roughness by $\sqrt(8 \log 2)$
    which is about 2.35.

    This interface can be computed analytically. When computing
    the slab product of the reflectivity, scale the Fresnel coefficient
    $F$ by the interface function yielding:

    .. math:

        F = (k_i-k_{i+1})/(k_i+k_{i+1}) \exp(-2 k_i k_{i+1} \sigma_i^2)
    """
    @classmethod
    def as_fwhm(cls, *args, **kw):
        self = cls(*args, **kw)
        self._scale = 1/sqrt(8*log(2))
        return self

    def __init__(self, width=0, name="erf"):
        self._scale = 1
        self.width = Parameter.default(width, limits=(0, inf), name=name)

    def parameters(self):
        return {'width':self.width}

    def cdf(self, z):
        sigma = self.width.value * self._scale
        if sigma <= 0.0:
            return 1.*(z >= 0)
        else:
            return 0.5*(1 + erf(z/(sigma*sqrt(2))))

    def pdf(self, z):
        sigma = self.width.value * self._scale
        if sigma <= 0.0:
            return inf*(z == 0)
        else:
            return exp(z**2/(-2*sigma**2)) / sqrt(2*pi*sigma**2)

    def ppf(self, z):
        sigma = self.width.value * self._scale
        if sigma <= 0.0:
            return 0*z
        else:
            return sigma*sqrt(2)*erfinv(2*z-1)

class Linear(Interface):
    """
    Linear interface.

    *width*  full width of the interface

    *name* (string: "tanh")

    The linear profile has the form::

        CDF(z) = 2/w*z if |z|<w/2, 0 if z<-w/2, 1 otherwise
        PDF(z) = 1/w if |z|<w/2, otherwise 0
        PPF(z) = w/2*z if |z|<w/2, -w/2 if z<-w/2, w/2 otherwise
    """
    def __init__(self, width=0, name="linear"):
        self.width = Parameter.default(width, limits=(0, inf), name=name)
    def parameters(self):
        return {'width':self.width}
    def cdf(self, z):
        w = float(self.width.value)
        if w <= 0.0:
            return 1.*(z >= 0)
        else:
            return np.clip(z/w + 0.5, 0, 1)
    def pdf(self, z):
        w = float(self.width.value)
        if w <= 0.0:
            return inf*(z == 0)
        else:
            return (abs(z) < w/2)/w
    def ppf(self, z):
        w = float(self.width.value)
        if w <= 0.0:
            return 0*z
        else:
            return np.clip(2*z/w, -w/2, w/2)

class Tanh(Interface):
    r"""
    Hyperbolic tangent profile

    *width* (Parameter: 0 Angstroms)

        1-\ $\sigma$ equivalent roughness.  For roughness $w$ measured by the
        full width at half maximum (FWHM), use Tanh.as_fwhm(w).

    *name* (string: "tanh")

    The tanh profile has the form:

    .. math:

        \text{CDF}(z) = (1 + \tanh(C/wz))/2
        \text{PDF}(z) = C/(2w) \sech((C/w)z)^2
        \text{PPF}(z) = (w/C) \tanh^{-1}(2z-1)

    where $w$ is the interface roughness and $C$ is a scaling constant.
    $C$ is $\tanh^{-1}(\text{erf}(1/\sqrt{2}))$ for width defined
    by 1-\ $\sigma$, or $C$ is $2\cosh^{-1}(\sqrt{2})$ for width
    defined by FWHM.

    .. Note::
       This profile was derived from the free energy of a nonuniform system:

        * J.W. Cahn and J. E. Hilliard, J. Chem. Phys. 28, 258 (1958)

    .. seealso::
       This profile has an analytic solution. E.S. Wu, and W. W. Webb,
       Phys Rev A 8(4) 2065-2076 (1973)
    """

    # Derivation
    # ==========
    #
    # To find C where w is defined as 1-sigma equivalent, use the
    # identity Erf.CDF(z=sigma;w=sigma) = tanh.CDF(z=sigma;w=sigma).
    # This simplifies to::
    #
    #    erf.CDF  = (1+erf(z/(w*sqrt(2)))/2 = (1+erf(1/sqrt(2)))/2
    #    tanh.CDF = (1+tanh(C/w*z))/2       = (1+tanh(C))/2
    #
    #    erf.CDF = tanh.CDF => C = atanh(erf(1/sqrt(2)))
    #
    # To find C where w is defined as FWHM, use the equivalent probability
    # density function::
    #
    #    PDF(z) = C/2w * sech(C/w*z)**2
    #
    # Solving PDF(w/2) = PDF(0)/2 yields::
    #
    #    Pw = PDF(w/2) = C/2w * sech(C/2)**2
    #    Po = PDF(0) = C/2w * sech(0)**2/2 = C/2w
    #
    #    Pw = Po/2 => sech(C/2)**2 = 1/2
    #              => C = 2 acosh(sqrt(2))
    #
    # To find ws 1-sigma given tanh fwhm of w, use the scale factor
    # s = C_1_sigma/C_fwhm = 1/2 atanh(erf(1/sqrt(2)))/acosh(sqrt(2))
    # to form ws = w*s
    C = atanh(erf(1/sqrt(2)))
    Cfwhm = 2*acosh(sqrt(2))
    @classmethod
    def as_fwhm(cls, *args, **kw):
        r"""
        Defines interface using FWHM rather than 1-\ $\sigma$.
        """
        self = cls(*args, **kw)
        self._scale = Tanh.C/Tanh.Cfwhm
        return self
    def __init__(self, width=0, name="tanh"):
        self._scale = 1
        self.width = Parameter.default(width, limits=(0, inf), name=name)
    def parameters(self):
        return {'width':self.width}
    def cdf(self, z):
        w = self.width.value * self._scale
        if w <= 0.0:
            return 1.*(z > 0)
        else:
            return 0.5*(1 + tanh((Tanh.C/w)*z))
    def pdf(self, z):
        w = self.width.value * self._scale
        if w <= 0.0:
            return inf*(z == 0)
        else:
            return sech((Tanh.C/w)*z)**2 * (Tanh.C / (2*w))
    def ppf(self, z):
        w = self.width.value * self._scale
        if w <= 0.0:
            return 0*z
        else:
            return (w/Tanh.C)*atanh(2*z-1)

def demo_fwhm():
    """
    Show the available interface functions and the corresponding probability
    density functions.
    """

    # Plot the cdf and pdf
    import matplotlib.pyplot as plt
    w = 10
    perf = Erf.as_fwhm(w)
    ptanh = Tanh.as_fwhm(w)

    z = plt.linspace(-w, w, 800)
    plt.subplot(211)
    plt.plot(z, perf.cdf(z))
    plt.plot(z, ptanh.cdf(z))
    plt.legend(['erf', 'tanh'])
    plt.grid(True)
    plt.subplot(212)
    plt.plot(z, perf.pdf(z), 'b')
    plt.plot(z, ptanh.pdf(z), 'g')
    plt.legend(['erf', 'tanh'])

    # Show fwhm
    arrowprops = dict(arrowstyle='wedge', connectionstyle='arc3', fc='0.6')
    bbox = dict(boxstyle='round', fc='0.8')
    plt.annotate('erf FWHM', xy=(w/2, perf.pdf(0)/2),
                 xytext=(-35, 10), textcoords="offset points",
                 arrowprops=arrowprops, bbox=bbox)
    plt.annotate('tanh FWHM', xy=(w/2, ptanh.pdf(0)/2),
                 xytext=(-35, -35), textcoords="offset points",
                 arrowprops=arrowprops, bbox=bbox)

    plt.grid(True)


def demo_tanh_to_erf():
    """
    Show the available interface functions and the corresponding probability
    density functions.
    """

    # Plot the cdf and pdf
    import matplotlib.pyplot as plt
    w = 10
    ws = w * Tanh.C/Tanh.Cfwhm
    ptanh = Tanh.as_fwhm(w)
    perf = Erf(ws)

    z = plt.linspace(-2*w, 2*w, 800)
    plt.subplot(211)
    plt.plot(z, perf.cdf(z))
    plt.plot(z, ptanh.cdf(z))
    plt.title("""FWHM tanh -> 1-sigma erf
scale by atanh(erf(1/sqrt(2))) / (2 acosh(sqrt(2)))""")
    plt.legend(['erf', 'tanh'])
    plt.grid(True)
    plt.subplot(212)
    plt.plot(z, perf.pdf(z), 'b')
    plt.plot(z, ptanh.pdf(z), 'g')
    plt.legend(['erf', 'tanh'])

    # Show fwhm
    arrowprops = dict(arrowstyle='wedge', connectionstyle='arc3', fc='0.6')
    bbox = dict(boxstyle='round', fc='0.8')
    plt.annotate('erf 1-sigma', xy=(ws, perf.pdf(ws)),
                 xytext=(-2, 20), textcoords="offset points",
                 arrowprops=arrowprops, bbox=bbox)
    plt.annotate('tanh FWHM', xy=(w/2, ptanh.pdf(0)/2),
                 xytext=(-58, -35), textcoords="offset points",
                 arrowprops=arrowprops, bbox=bbox)

    plt.grid(True)

def demo():
    """
    Show the available interface functions and the corresponding probability
    density functions.
    """

    # Plot the cdf and pdf
    import matplotlib.pyplot as plt
    w = 10
    perf = Erf(w)
    ptanh = Tanh(w)
    plinear = Linear(2.35*w)

    #arrowprops=dict(arrowstyle='wedge', connectionstyle='arc3', fc='0.6')
    #bbox=dict(boxstyle='round', fc='0.8')

    z = plt.linspace(-3*w, 3*w, 800)
    plt.subplot(211)
    plt.plot(z, perf.cdf(z))
    plt.plot(z, ptanh.cdf(z))
    plt.plot(z, plinear.cdf(z))
    plt.axvline(w, linewidth=2)
    plt.annotate('1-sigma', xy=(w*1.1, 0.2))
    plt.legend(['erf', 'tanh'])
    plt.grid(True)
    plt.subplot(212)
    plt.plot(z, perf.pdf(z))
    plt.plot(z, ptanh.pdf(z))
    plt.plot(z, plinear.pdf(z))
    plt.axvline(w, linewidth=2)
    plt.annotate('1-sigma', xy=(w*1.1, 0.2))
    plt.legend(['erf', 'tanh', 'linear'])
    plt.grid(True)

def _test_one(name, p, w, tol):
    import scipy.integrate as sum
    # Check that the pdf approximately matchs the numerical integral
    # Check that integral(-inf, 0) of pdf sums to 0.5
    err = abs(sum.romberg(p.pdf, -20*w, 0, tol=1e-15) - 0.5)
    assert err < tol, "%s cdf(0) == 0.5 yields %g"%(name, err)

    # Check that integral(-inf, x) of pdf sums to cdf when x != 0
    err = abs(sum.romberg(p.pdf, -20*w, w/6, tol=1e-15) - p.cdf(w/6))
    assert err < tol, "%s cdf(w/6) == w/6 yields %g"%(name, err)

    # Check that P = cdf(ppf(P))
    P = 0.002
    err = abs(p.cdf(p.ppf(P)) - P)
    assert err < tol, "%s p(lo) = P yields %g"%(name, err)

def test():
    w = 1.5
    _test_one('Erf', Erf(w), w=w, tol=1e-13)
    _test_one('Tanh', Tanh(w), w=w, tol=1e-11)
    _test_one('Erf:fwhm', Erf.as_fwhm(2.35*w), w=w, tol=1e-11)
    _test_one('Tanh:fwhm', Tanh.as_fwhm(2*w), w=w, tol=1e-11)

if __name__ == "__main__":
    #demo()
    #demo_fwhm()
    #demo_tanh_to_erf()
    test()



#!/usr/bin/env python
"""
Run bumps with refl1d plugin
"""
from . import __version__

def setup_bumps():
    """
    Install the refl1d plugin into bumps, but don't run main.
    """
    # vfs_init must be called pretty much first since it replaces filesystem
    # calls with filesystem hooks.  Any module that imports the calls directly
    # e.g., using *from os.path import exists*, needs to have the hook in
    # place before the module is even imported.  This includes
    try:
        from bumps.vfs import vfs_init
        vfs_init()
    except ImportError:
        # CRUFT: older bumps doesn't provide vfs
        pass
    import bumps.cli
    bumps.cli.set_mplconfig(appdatadir='Refl1D-'+__version__)
    from . import fitplugin
    bumps.cli.install_plugin(fitplugin)

def cli():
    """
    Install the Refl1D plugin into bumps and run the command line interface.
    """
    setup_bumps()
    import bumps.cli
    bumps.cli.main()

def gui():
    """
    Install the Refl1D plugin into bumps and run the graphical user interface.
    """
    setup_bumps()
    import bumps.gui.gui_app
    bumps.gui.gui_app.main()

if __name__ == "__main__":
    cli()



"""
Freeform modeling with B-Splines
"""

import numpy as np
from numpy import inf
from bumps.parameter import Parameter as Par
from bumps.bspline import pbs, bspline

from .model import Layer
from . import util

#TODO: add left_sld, right_sld to all layers so that fresnel works
#TODO: access left_sld, right_sld so freeform doesn't need left, right
#TODO: restructure to use vector parameters
#TODO: allow the number of layers to be adjusted by the fit
class FreeLayer(Layer):
    """
    A freeform section of the sample modeled with B-splines.

    sld (rho) and imaginary sld (irho) can be modeled with a separate
    number of control points. The control points can be equally spaced
    in the layers unless rhoz or irhoz are specified. If the z values
    are given, they must be in the range [0, 1].  One control point is
    anchored at either end, so there are two fewer z values than controls
    if z values are given.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """
    def __init__(self, thickness=0, left=None, right=None,
                 rho=(), irho=(), rhoz=(), irhoz=(), name="Freeform"):
        self.name = name
        self.left, self.right = left, right
        self.thickness = Par.default(thickness, limits=(0, inf),
                                     name=name+" thickness")
        self.rho, self.irho, self.rhoz, self.irhoz \
            = [[Par.default(p, name=name+" [%d] %s"%(i, part), limits=limits)
                for i, p in enumerate(v)]
               for v, part, limits
               in zip((rho, irho, rhoz, irhoz),
                      ('rho', 'irho', 'rhoz', 'irhoz'),
                      ((-inf, inf), (-inf, inf), (0, 1), (0, 1)))
              ]
        if len(self.rhoz) > 0 and len(self.rhoz) != len(self.rho):
            raise ValueError("must have one z value for each rho")
        if len(self.irhoz) > 0 and len(self.irhoz) != len(self.irho):
            raise ValueError("must have one z value for each irho")
    def parameters(self):
        return {'rho':self.rho,
                'rhoz':self.rhoz,
                'irho':self.irho,
                'irhoz':self.irhoz,
                'left':self.left.parameters(),
                'right':self.right.parameters(),
               }
    def render(self, probe, slabs):
        thickness = self.thickness.value
        left_rho, left_irho = self.left.sld(probe)
        right_rho, right_irho = self.right.sld(probe)
        Pw, Pz = slabs.microslabs(thickness)
        t = Pz/thickness
        Prho = _profile(left_rho, right_rho, self.rho, self.rhoz, t)
        Pirho = _profile(left_irho, right_irho, self.irho, self.irhoz, t)
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)

class FreeformInterface01(Layer):
    """
    A freeform section of the sample modeled with B-splines.

    sld (rho) and imaginary sld (irho) can be modeled with a separate
    number of control points. The control points can be equally spaced
    in the layers unless rhoz or irhoz are specified. If the z values
    are given, they must be in the range [0, 1].  One control point is
    anchored at either end, so there are two fewer z values than controls
    if z values are given.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """
    def __init__(self, thickness=0, interface=0,
                 below=None, above=None,
                 z=None, vf=None, name="Interface"):
        self.name = name
        self.below, self.above = below, above
        self.thickness = Par.default(thickness, limits=(0, inf),
                                     name=name+" thickness")
        self.interface = Par.default(interface, limits=(0, inf),
                                     name=name+" interface")
        if len(z) != len(vf):
            raise ValueError("Need one vf for every z")
        #if len(z) != len(vf)+2:
        #    raise ValueError("Only need vf for interior z, so len(z)=len(vf)+2")
        self.z = [Par.default(p, name=name+" z[%d]"%i, limits=(0, 1))
                  for i, p in enumerate(z)]
        self.vf = [Par.default(p, name=name+" vf[%d]"%i, limits=(0, 1))
                   for i, p in enumerate(vf)]
    def parameters(self):
        return {'z':self.z,
                'vf':self.vf,
                'below':self.below.parameters(),
                'above':self.above.parameters(),
               }
    def render(self, probe, slabs):
        thickness = self.thickness.value
        left_rho, left_irho = self.below.sld(probe)
        right_rho, right_irho = self.above.sld(probe)
        z = np.hstack((0, sorted([v.value for v in self.z]), 1))
        vf = np.hstack((0, sorted([v.value for v in self.vf]), 1))
        Pw, Pz = slabs.microslabs(thickness)
        t = Pz/thickness
        offset, profile = pbs(z, vf, t, parametric=False, clamp=True)
        Pw, profile = util.merge_ends(Pw, profile, tol=1e-3)
        Prho = (1-profile)*left_rho + profile*right_rho
        Pirho = (1-profile)*left_irho + profile*right_irho
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)

class FreeInterface(Layer):
    """
    A freeform section of the sample modeled with monotonic splines.

    Layers have a slope of zero at the ends, so the automatically blend
    with slabs.
    """
    def __init__(self, interface=0,
                 below=None, above=None,
                 dz=None, dp=None, name="Interface"):
        self.name = name
        self.below, self.above = below, above
        self.interface = Par.default(interface, limits=(0, inf),
                                     name=name+" interface")

        # Choose reasonable defaults if not given
        if dp is None and dz is None:
            dp = [1]*5
        if dp is None:
            dp = [1]*len(dz)
        if dz is None:
            dz = [10./len(dp)]*len(dp)
        if len(dz) != len(dp):
            raise ValueError("Need one dz for every dp")
        #if len(z) != len(vf)+2:
        #    raise ValueError("Only need vf for interior z, so len(z)=len(vf)+2")
        self.dz = [Par.default(p, name=name+" dz[%d]"%i, limits=(0, inf))
                   for i, p in enumerate(dz)]
        self.dp = [Par.default(p, name=name+" dp[%d]"%i, limits=(0, inf))
                   for i, p in enumerate(dp)]
    def _get_thickness(self):
        w = sum(v.value for v in self.dz)
        return Par(w, name=self.name+" thickness")
    def _set_thickness(self, v):
        if v != 0:
            raise ValueError("thickness cannot be set for FreeInterface")
    thickness = property(_get_thickness, _set_thickness)

    def parameters(self):
        return {'dz':self.dz,
                'dp':self.dp,
                'below':self.below.parameters(),
                'above':self.above.parameters(),
               }
    def render(self, probe, slabs):
        left_rho, left_irho = self.below.sld(probe)
        right_rho, right_irho = self.above.sld(probe)
        z = np.hstack((0, np.cumsum([v.value for v in self.dz])))
        p = np.hstack((0, np.cumsum([v.value for v in self.dp])))
        if p[-1] == 0:
            p[-1] = 1
        p /= p[-1]
        Pw, Pz = slabs.microslabs(z[-1])
        _, profile = pbs(z, p, Pz, parametric=False, clamp=True)
        profile = np.clip(profile, 0, 1)
        Pw, profile = util.merge_ends(Pw, profile, tol=1e-3)
        Prho = (1-profile)*left_rho + profile*right_rho
        Pirho = (1-profile)*left_irho + profile*right_irho
        slabs.extend(rho=[Prho], irho=[Pirho], w=Pw)

def _profile(left, right, control, z, t):
    cy = np.hstack((left, [p.value for p in control], right))
    if len(z) > 0:
        cx = np.hstack((0, [p.value for p in z], 1))
        return pbs(cx, cy, t)
    else:
        return bspline(cy, t)



from __future__ import with_statement

import wx
# Can't seem to detect when notebook should be drawn on Mac
IS_MAC = (wx.Platform == '__WXMAC__')

from numpy import inf

from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigureCanvas
from matplotlib.backends.backend_wxagg import NavigationToolbar2WxAgg as Toolbar

# The Figure object is used to create backend-independent plot representations.
from matplotlib.figure import Figure

import matplotlib.pyplot as plt

from bumps.gui.util import EmbeddedPylab
from bumps.fitproblem import MultiFitProblem

from refl1d.probe import Probe


# ------------------------------------------------------------------------
class DataView(wx.Panel):
    title = 'Reflectivity'
    default_size = (600,400)
    def __init__(self, *args, **kw):
        wx.Panel.__init__(self, *args, **kw)

        # Instantiate a figure object that will contain our plots.
        figure = Figure(figsize=(1,1), dpi=72)

        # Initialize the figure canvas, mapping the figure object to the plot
        # engine backend.
        canvas = FigureCanvas(self, wx.ID_ANY, figure)

        # Wx-Pylab magic ...
        # Make our canvas an active figure manager for pylab so that when
        # pylab plotting statements are executed they will operate on our
        # canvas and not create a new frame and canvas for display purposes.
        # This technique allows this application to execute code that uses
        # pylab stataments to generate plots and embed these plots in our
        # application window(s).  Use _activate_figure() to set.
        self.pylab_interface = EmbeddedPylab(canvas)

        # Instantiate the matplotlib navigation toolbar and explicitly show it.
        mpl_toolbar = Toolbar(canvas)
        mpl_toolbar.Realize()

        # Create a vertical box sizer to manage the widgets in the main panel.
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(canvas, 1, wx.EXPAND|wx.LEFT|wx.RIGHT, border=0)
        sizer.Add(mpl_toolbar, 0, wx.EXPAND|wx.ALL, border=0)

        # Associate the sizer with its container.
        self.SetSizer(sizer)
        sizer.Fit(self)

        self.view = Probe.view

        self._need_redraw = False
        self.Bind(wx.EVT_SHOW, self.OnShow)
        self._calculating = False
        self.toolbar = mpl_toolbar

    def menu(self):
        # Add 'View' menu to the menu bar and define its options.
        # Present y-axis plotting scales as radio buttons.
        # Grey out items that are not currently implemented.
        frame = wx.GetTopLevelParent(self)
        menu = wx.Menu()
        _item = menu.AppendRadioItem(wx.ID_ANY,
                                          "&Fresnel",
                                          "Plot R/R_F")
        frame.Bind(wx.EVT_MENU, self.OnFresnel, _item)
        _item.Check(Probe.view == 'fresnel')
        _item = menu.AppendRadioItem(wx.ID_ANY,
                                          "Log Fresnel",
                                          "Plot log R/R_F")
        frame.Bind(wx.EVT_MENU, self.OnLogFresnel, _item)
        _item.Check(Probe.view == 'logfresnel')
        _item = menu.AppendRadioItem(wx.ID_ANY,
                                          "Li&near",
                                          "Plot linear R")
        frame.Bind(wx.EVT_MENU, self.OnLinear, _item)
        _item.Check(Probe.view == 'linear')
        _item = menu.AppendRadioItem(wx.ID_ANY,
                                          "&Log",
                                          "Plot log R")
        frame.Bind(wx.EVT_MENU, self.OnLog, _item)
        _item.Check(Probe.view == 'log')
        _item = menu.AppendRadioItem(wx.ID_ANY,
                                          "&Q4",
                                          "Plot R * Q^4")
        frame.Bind(wx.EVT_MENU, self.OnQ4, _item)
        _item.Check(Probe.view == 'q4')
        _item = menu.AppendRadioItem(wx.ID_ANY,
                                          "&SA",
                                          "Plot spin asymmetry")
        frame.Bind(wx.EVT_MENU, self.OnSA, _item)
        _item.Check(Probe.view == 'SA')

        menu.AppendSeparator()

        _item = menu.Append(wx.ID_ANY,
                                 "&Residuals",
                                 "Plot residuals (R_theory - R)/dR")
        frame.Bind(wx.EVT_MENU, self.OnResiduals, _item)
        menu.Enable(id=_item.GetId(), enable=True)

        return menu

    # ==== Views ====
    def OnLog(self, event):
        self.view = "log"
        self.redraw()

    def OnLinear(self, event):
        self.view = "linear"
        self.redraw()

    def OnFresnel(self, event):
        self.view = "fresnel"
        self.redraw()

    def OnLogFresnel(self, event):
        self.view = "logfresnel"
        self.redraw()

    def OnQ4(self, event):
        self.view = "q4"
        self.redraw()

    def OnSA(self, event):
        self.view = "SA"
        self.redraw()

    def OnResiduals(self, event):
        self.view = "residuals"
        self.redraw()

    # ==== Model view interface ===
    def OnShow(self, event):
        #print "theory show"
        if not event.Show: return
        #print "showing theory"
        if self._need_redraw:
            #print "-redraw"
            self.redraw()

    def get_state(self):
        return self.problem

    def set_state(self, state):
        self.set_model(state)

    def set_model(self, model):
        # print ">>>>>>> refl1d data set model"
        self.problem = model
        self.redraw(reset=True)

    def update_model(self, model):
        # print ">>>>>>> refl1d data update model"
        if self.problem == model:
            self.redraw()

    def update_parameters(self, model):
        # print ">>>>>>> refl1d data update parameters"
        if self.problem == model:
            self.redraw()
    # =============================

    def redraw(self, reset=False):
        # Hold off drawing until the tab is visible
        if not IS_MAC and not self.IsShown():
            self._need_redraw = True
            return
        #print "drawing theory"

        if self._calculating:
            # That means that I've entered the thread through a
            # wx.Yield for the currently executing redraw.  I need
            # to cancel the running thread and force it to start
            # the calculation over.
            self._cancel_calculate = True
            #print "canceling calculation"
            return

        self._need_redraw = False
        self._calculating = True

        # Calculate reflectivity
        #print "calling again"
        while True:
            #print "restarting"
            # We are restarting the calculation, so clear the reset flag
            self._cancel_calculate = False

            # Preform the calculation
            if isinstance(self.problem,MultiFitProblem):
                #print "n=",len(self.problem.models)
                for p in self.problem.models:
                    if hasattr(p.fitness, 'reflectivity'):
                        self._precalc(p.fitness)
                    #print "cancel",self._cancel_calculate,"reset",p.fitness.is_reset()
                        if p.fitness.is_reset() or self._cancel_calculate: break
                if self._cancel_calculate \
                    or self.problem.active_model.fitness.is_reset(): continue
            else:
                self._precalc(self.problem.fitness)
                if self._cancel_calculate \
                    or self.problem.fitness.is_reset(): continue

            # Redraw the canvas with newly calculated reflectivity
            with self.pylab_interface:
                ax = plt.gca()
                #print "reset",reset, ax.get_autoscalex_on(), ax.get_xlim()
                reset = reset or ax.get_autoscalex_on()
                range_x = ax.get_xlim()
                #print "composing"
                plt.clf() # clear the canvas
                #shift=20 if self.view == 'log' else 0
                shift=0
                if isinstance(self.problem,MultiFitProblem):
                    for _,p in enumerate(self.problem.models):
                        if hasattr(p.fitness, 'reflectivity'):
                            p.fitness.plot_reflectivity(view=self.view,
                                                        plot_shift=shift)
                            if self._cancel_calculate or p.fitness.is_reset(): break
                    if self._cancel_calculate \
                        or self.problem.active_model.fitness.is_reset(): continue
                else:
                    self.problem.fitness.plot_reflectivity(view=self.view,
                                                           plot_shift=shift)
                    if self._cancel_calculate \
                        or self.problem.fitness.is_reset(): continue

                try:
                    # If we can calculate chisq, then put it on the graph.
                    text = "chisq=%g"%self.problem.chisq()
                    constraints = self.problem.parameter_nllf() + self.problem.constraints_nllf()
                    if constraints > 0: text+= " constraints=%g"%constraints
                    plt.text(0.01, 0.01, text, transform=plt.gca().transAxes)
                except Exception:
                    pass
                #print "drawing"
                if not reset:
                    self.toolbar.push_current()
                    set_xrange(plt.gca(), range_x)
                    self.toolbar.push_current()
                plt.draw()
                #print "done drawing"
                break

        self._calculating = False

    def _precalc(self, fitness):
        # First calculate reflectivity
        fitness.reflectivity(resolution=False)
        #print "yield 1"
        wx.Yield()
        if self._cancel_calculate or fitness.is_reset(): return
        # Then calculate resolution
        fitness.reflectivity()
        #print "yield 2"
        wx.Yield()
        if self._cancel_calculate or fitness.is_reset(): return

def set_xrange(ax, range_x):
    miny,maxy = inf,-inf
    for L in ax.get_lines():
        x,y = L.get_data()
        idx = (x>range_x[0]) & (x<range_x[1])
        if idx.any():
            miny = min(miny,min(y[idx]))
            maxy = max(maxy,max(y[idx]))
    if miny < maxy:
        if ax.get_yscale() == 'linear':
            padding = 0.05*(maxy-miny)
            miny,maxy = miny-padding, maxy+padding
        else:
            miny,maxy = miny*0.95, maxy*1.05
    ax.set_xlim(range_x)
    ax.set_ylim(miny,maxy)
